from mhd_model.model.definitions import MHD_MODEL_V0_1_DEFAULT_SCHEMA_NAME
from mhd_model.model.v0_1.dataset.validation.profile.base import (
    EmbeddedRefValidation,
    RelationshipValidation,
)
from mhd_model.model.v0_1.dataset.validation.profile.definition import (
    CvNodeValidation,
    CvTermValidation,
    FilterCondition,
    MhDatasetValidation,
    NodePropertyValidation,
    NodeValidation,
    PropertyConstraint,
)
from mhd_model.model.v0_1.rules.managed_cv_terms import (
    COMMON_ASSAY_TYPES,
    COMMON_CHARACTERISTIC_DEFINITIONS,
    COMMON_MEASUREMENT_TYPES,
    COMMON_OMICS_TYPES,
    COMMON_PARAMETER_DEFINITIONS,
    COMMON_PROTOCOL_PARAMETERS,
    COMMON_PROTOCOLS,
    COMMON_TECHNOLOGY_TYPES,
)
from mhd_model.shared.model import CvTerm
from mhd_model.shared.validation.definitions import (
    AllowAnyCvTerm,
    AllowedChildrenCvTerms,
    AllowedCvList,
    AllowedCvTerms,
    CvTermPlaceholder,
    ParentCvTerm,
)

MHD_BASE_VALIDATION_V0_1 = MhDatasetValidation(
    schema=MHD_MODEL_V0_1_DEFAULT_SCHEMA_NAME
)

MHD_BASE_VALIDATION_V0_1.mhd_nodes = [
    NodeValidation(
        node_type="assay",
        min=0,
        validations=[
            NodePropertyValidation(
                node_type="assay",
                node_property_name="repository_identifier",
                contraints=PropertyConstraint(required=True, min_length=2),
            ),
            NodePropertyValidation(
                node_type="assay",
                node_property_name="name",
                contraints=PropertyConstraint(required=True, min_length=2),
            ),
            EmbeddedRefValidation(
                node_type="assay",
                node_property_name="metadata_file_ref",
                required=False,
                target_ref_types=["metadata-file"],
            ),
            EmbeddedRefValidation(
                node_type="assay",
                node_property_name="technology_type_ref",
                required=False,
                target_ref_types=["descriptor"],
            ),
            EmbeddedRefValidation(
                node_type="assay",
                node_property_name="assay_type_ref",
                required=False,
                target_ref_types=["descriptor"],
            ),
            EmbeddedRefValidation(
                node_type="assay",
                node_property_name="measurement_type_ref",
                required=False,
                target_ref_types=["descriptor"],
            ),
            EmbeddedRefValidation(
                node_type="assay",
                node_property_name="omics_type_ref",
                required=False,
                target_ref_types=["descriptor"],
            ),
            EmbeddedRefValidation(
                node_type="assay",
                node_property_name="protocol_refs",
                required=False,
                target_ref_types=["protocol"],
            ),
            EmbeddedRefValidation(
                node_type="assay",
                node_property_name="sample_run_refs",
                required=False,
                target_ref_types=["sample-run"],
            ),
        ],
        relationships=[
            RelationshipValidation(
                description="A link to a descriptor that describes the assay.",
                source="assay",
                relationship_name="described-as",
                reverse_relationship_name="describes",
                target="descriptor",
                min=0,
                min_for_each_source=0,
                max_for_each_source=0,
            ),
            RelationshipValidation(
                description="A link to a study that the assay was conducted as part of it "
                "to generate data addressing study objectives",
                source="assay",
                relationship_name="part-of",
                reverse_relationship_name="has-assay",
                target="study",
                min=0,
                min_for_each_source=0,
                max_for_each_source=0,
            ),
            RelationshipValidation(
                description="A link to a protocol conducted in assay.",
                source="assay",
                relationship_name="follows",
                reverse_relationship_name="used-in",
                target="protocol",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
    NodeValidation(
        node_type="characteristic-definition",
        min=0,
        validations=[
            EmbeddedRefValidation(
                node_type="characteristic-definition",
                node_property_name="characteristic_type_ref",
                required=True,
                target_ref_types=["characteristic-type"],
            ),
            NodePropertyValidation(
                node_type="characteristic-definition",
                node_property_name="name",
                contraints=PropertyConstraint(required=True, min_length=2),
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="characteristic-definition",
                relationship_name="has-instance",
                reverse_relationship_name="instance-of",
                target="characteristic-value",
                min=2,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="characteristic-definition",
                relationship_name="has-type",
                reverse_relationship_name="type-of",
                target="characteristic-type",
                min=0,
                min_for_each_source=1,
                max_for_each_source=1,
            ),
            RelationshipValidation(
                source="characteristic-definition",
                relationship_name="used-in",
                reverse_relationship_name="has-characteristic-definition",
                target="study",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
    NodeValidation(
        node_type="derived-data-file",
        min=0,
        validations=[
            NodePropertyValidation(
                node_type="derived-data-file",
                node_property_name="name",
                contraints=PropertyConstraint(required=True, min_length=2),
            ),
            NodePropertyValidation(
                node_type="derived-data-file",
                node_property_name="format_ref",
                contraints=PropertyConstraint(required=True),
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="derived-data-file",
                relationship_name="referenced-in",
                reverse_relationship_name="references",
                target="metadata-file",
                min=0,
                min_for_each_source=0,
                max_for_each_source=0,
            ),
            RelationshipValidation(
                source="derived-data-file",
                relationship_name="created-in",
                reverse_relationship_name="has-derived-data-file",
                target="study",
                min=0,
                min_for_each_source=1,
                max_for_each_source=1,
            ),
            RelationshipValidation(
                description="A link to a descriptor that describes the derived data file.",
                source="derived-data-file",
                relationship_name="described-as",
                reverse_relationship_name="describes",
                target="descriptor",
                min=0,
                min_for_each_source=0,
                max_for_each_source=0,
            ),
        ],
    ),
    NodeValidation(
        node_type="factor-definition",
        min=0,
        validations=[
            EmbeddedRefValidation(
                node_type="factor-definition",
                node_property_name="factor_type_ref",
                required=True,
                target_ref_types=["factor-type"],
            ),
            NodePropertyValidation(
                node_type="factor-definition",
                node_property_name="name",
                contraints=PropertyConstraint(required=True, min_length=2),
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="factor-definition",
                relationship_name="has-type",
                reverse_relationship_name="type-of",
                target="factor-type",
                min=0,
                min_for_each_source=0,
                max_for_each_source=0,
            ),
            RelationshipValidation(
                source="factor-definition",
                relationship_name="has-instance",
                reverse_relationship_name="instance-of",
                target="factor-value",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="factor-definition",
                relationship_name="used-in",
                reverse_relationship_name="has-factor-definition",
                target="study",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
    NodeValidation(
        node_type="metadata-file",
        min=0,
        validations=[
            NodePropertyValidation(
                node_type="metadata-file",
                node_property_name="url_list",
                contraints=PropertyConstraint(required=True, min_length=1),
            ),
            NodePropertyValidation(
                node_type="metadata-file",
                node_property_name="name",
                contraints=PropertyConstraint(required=True, min_length=2),
            ),
            NodePropertyValidation(
                node_type="metadata-file",
                node_property_name="format_ref",
                contraints=PropertyConstraint(required=True),
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="metadata-file",
                relationship_name="described-as",
                reverse_relationship_name="describes",
                target="descriptor",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="metadata-file",
                relationship_name="referenced-in",
                reverse_relationship_name="references",
                target="metadata-file",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="metadata-file",
                relationship_name="describes",
                reverse_relationship_name="has-metadata-file",
                target="study",
                min=0,
                min_for_each_source=1,
                max_for_each_source=1,
            ),
            RelationshipValidation(
                source="metadata-file",
                relationship_name="references",
                reverse_relationship_name="referenced-in",
                target="derived-data-file",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="metadata-file",
                relationship_name="references",
                reverse_relationship_name="referenced-in",
                target="raw-data-file",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="metadata-file",
                relationship_name="references",
                reverse_relationship_name="referenced-in",
                target="result-file",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="metadata-file",
                relationship_name="references",
                reverse_relationship_name="referenced-in",
                target="supplementary-file",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
    NodeValidation(
        node_type="metabolite",
        min=0,
        validations=[
            NodePropertyValidation(
                node_type="metabolite",
                node_property_name="name",
                contraints=PropertyConstraint(required=True, min_length=2),
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="metabolite",
                relationship_name="identified-as",
                reverse_relationship_name="reported-identifier-of",
                target="metabolite-identifier",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="metabolite",
                relationship_name="described-as",
                reverse_relationship_name="describes",
                target="descriptor",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="metabolite",
                relationship_name="reported-in",
                reverse_relationship_name="reports",
                target="study",
                min=0,
                min_for_each_source=1,
            ),
        ],
    ),
    NodeValidation(
        node_type="organization",
        min=0,
        validations=[
            NodePropertyValidation(
                node_type="organization",
                node_property_name="name",
                contraints=PropertyConstraint(required=True, min_length=10),
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="organization",
                relationship_name="funds",
                reverse_relationship_name="funded-by",
                target="project",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="organization",
                relationship_name="funds",
                reverse_relationship_name="funded-by",
                target="study",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="organization",
                relationship_name="manages",
                reverse_relationship_name="managed-by",
                target="project",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="organization",
                relationship_name="coordinates",
                reverse_relationship_name="coordinated-by",
                target="project",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="organization",
                relationship_name="affiliates",
                reverse_relationship_name="affiliated-with",
                target="person",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="organization",
                relationship_name="described-as",
                reverse_relationship_name="describes",
                target="descriptor",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
    NodeValidation(
        node_type="parameter-definition",
        min=0,
        validations=[
            EmbeddedRefValidation(
                node_type="parameter-definition",
                node_property_name="parameter_type_ref",
                required=True,
                target_ref_types=["parameter-type"],
            ),
            NodePropertyValidation(
                node_type="parameter-definition",
                node_property_name="name",
                contraints=PropertyConstraint(required=True, min_length=2),
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="parameter-definition",
                relationship_name="has-instance",
                reverse_relationship_name="instance-of",
                target="parameter-value",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="parameter-definition",
                relationship_name="has-type",
                reverse_relationship_name="type-of",
                target="parameter-type",
                min=0,
                min_for_each_source=1,
                max_for_each_source=1,
            ),
            RelationshipValidation(
                source="parameter-definition",
                relationship_name="used-in",
                reverse_relationship_name="has-parameter-definition",
                target="protocol",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
    NodeValidation(
        node_type="person",
        min=0,
        validations=[
            NodePropertyValidation(
                node_type="person",
                node_property_name="full_name",
                contraints=PropertyConstraint(required=True, min_length=5),
            ),
            NodePropertyValidation(
                node_type="person",
                node_property_name="emails",
                contraints=PropertyConstraint(required=True, min_length=1),
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="person",
                relationship_name="described-as",
                reverse_relationship_name="describes",
                target="descriptor",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="person",
                relationship_name="affiliated-with",
                reverse_relationship_name="affiliates",
                target="organization",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="person",
                relationship_name="contributes",
                reverse_relationship_name="has-contributor",
                target="project",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="person",
                relationship_name="contributes",
                reverse_relationship_name="has-contributor",
                target="study",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="person",
                relationship_name="principal-investigator-of",
                reverse_relationship_name="has-principal-investigator",
                target="study",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="person",
                relationship_name="submits",
                reverse_relationship_name="submitted-by",
                target="study",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="person",
                relationship_name="author-of",
                reverse_relationship_name="has-author",
                target="publication",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
    NodeValidation(
        node_type="project",
        min=0,
        validations=[
            NodePropertyValidation(
                node_type="project",
                node_property_name="title",
                contraints=PropertyConstraint(required=True, min_length=25),
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="project",
                relationship_name="described-as",
                reverse_relationship_name="describes",
                target="descriptor",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="project",
                relationship_name="funded-by",
                reverse_relationship_name="funds",
                target="organization",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="project",
                relationship_name="managed-by",
                reverse_relationship_name="manages",
                target="organization",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="project",
                relationship_name="coordinated-by",
                reverse_relationship_name="coordinates",
                target="organization",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="project",
                relationship_name="has-contributor",
                reverse_relationship_name="contributes",
                target="person",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="project",
                relationship_name="has-publication",
                reverse_relationship_name="describes",
                target="publication",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="project",
                relationship_name="has-study",
                reverse_relationship_name="part-of",
                target="study",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
    NodeValidation(
        node_type="protocol",
        min=0,
        validations=[
            NodePropertyValidation(
                node_type="protocol",
                node_property_name="name",
                contraints=PropertyConstraint(required=True),
            ),
            EmbeddedRefValidation(
                node_type="protocol",
                node_property_name="protocol_type_ref",
                required=True,
                target_ref_types=["protocol-type"],
            ),
            NodePropertyValidation(
                node_type="protocol",
                node_property_name="description",
                contraints=PropertyConstraint(required=False),
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="protocol",
                relationship_name="described-as",
                reverse_relationship_name="describes",
                target="descriptor",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="protocol",
                relationship_name="used-in",
                reverse_relationship_name="follows",
                target="assay",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="protocol",
                relationship_name="has-parameter-definition",
                reverse_relationship_name="used-in",
                target="parameter-definition",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="protocol",
                relationship_name="used-in",
                reverse_relationship_name="has-protocol",
                target="study",
                min=0,
                min_for_each_source=1,
            ),
            RelationshipValidation(
                source="protocol",
                relationship_name="has-type",
                reverse_relationship_name="type-of",
                target="protocol-type",
                min=0,
                min_for_each_source=1,
                max_for_each_source=1,
            ),
        ],
    ),
    NodeValidation(
        node_type="publication",
        min=0,
        validations=[
            NodePropertyValidation(
                node_type="publication",
                node_property_name="title",
                contraints=PropertyConstraint(required=True),
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="publication",
                relationship_name="described-as",
                reverse_relationship_name="describes",
                target="descriptor",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="publication",
                relationship_name="describes",
                reverse_relationship_name="has-publication",
                target="project",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="publication",
                relationship_name="describes",
                reverse_relationship_name="has-publication",
                target="study",
                min=0,
                min_for_each_source=0,
                max_for_each_source=1,
            ),
            RelationshipValidation(
                source="publication",
                relationship_name="has-author",
                reverse_relationship_name="author-of",
                target="person",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
    NodeValidation(
        node_type="raw-data-file",
        min=0,
        validations=[
            NodePropertyValidation(
                node_type="raw-data-file",
                node_property_name="url_list",
                contraints=PropertyConstraint(required=True, min_length=1),
            ),
            NodePropertyValidation(
                node_type="raw-data-file",
                node_property_name="name",
                contraints=PropertyConstraint(required=True, min_length=2),
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="raw-data-file",
                relationship_name="described-as",
                reverse_relationship_name="describes",
                target="descriptor",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="raw-data-file",
                relationship_name="created-in",
                reverse_relationship_name="has-raw-data-file",
                target="study",
                min=0,
                min_for_each_source=1,
            ),
            RelationshipValidation(
                source="raw-data-file",
                relationship_name="referenced-in",
                reverse_relationship_name="references",
                target="metadata-file",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
    NodeValidation(
        node_type="result-file",
        min=0,
        validations=[
            NodePropertyValidation(
                node_type="result-file",
                node_property_name="url_list",
                contraints=PropertyConstraint(required=True, min_length=1),
            ),
            NodePropertyValidation(
                node_type="result-file",
                node_property_name="name",
                contraints=PropertyConstraint(required=True, min_length=2),
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="result-file",
                relationship_name="described-as",
                reverse_relationship_name="describes",
                target="descriptor",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="result-file",
                relationship_name="created-in",
                reverse_relationship_name="has-result-file",
                target="study",
                min=0,
                min_for_each_source=1,
            ),
            RelationshipValidation(
                source="result-file",
                relationship_name="referenced-in",
                reverse_relationship_name="references",
                target="metadata-file",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
    NodeValidation(
        node_type="sample",
        min=0,
        validations=[
            NodePropertyValidation(
                node_type="sample",
                node_property_name="name",
                contraints=PropertyConstraint(required=True, min_length=1),
            ),
            NodePropertyValidation(
                node_type="sample",
                node_property_name="repository_identifier",
                contraints=PropertyConstraint(required=True),
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="sample",
                relationship_name="described-as",
                reverse_relationship_name="describes",
                target="descriptor",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="sample",
                relationship_name="has-factor-value",
                reverse_relationship_name="value-of",
                target="factor-value",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="sample",
                relationship_name="used-in",
                reverse_relationship_name="has-sample",
                target="study",
                min=0,
                min_for_each_source=1,
                max_for_each_source=1,
            ),
            RelationshipValidation(
                source="sample",
                relationship_name="derived-from",
                reverse_relationship_name="source-of",
                target="subject",
                min=0,
                min_for_each_source=1,
            ),
            RelationshipValidation(
                source="sample",
                relationship_name="derived-from",
                reverse_relationship_name="source-of",
                target="specimen",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="sample",
                relationship_name="has-characteristic-value",
                reverse_relationship_name="value-of",
                target="characteristic-value",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
    NodeValidation(
        node_type="sample-run",
        min=0,
        validations=[
            EmbeddedRefValidation(
                node_type="sample-run",
                node_property_name="sample_ref",
                required=True,
                target_ref_types=["sample"],
            ),
            EmbeddedRefValidation(
                node_type="sample-run",
                node_property_name="raw_data_file_refs",
                required=True,
                target_ref_types=["raw-data-file"],
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="sample-run",
                relationship_name="described-as",
                reverse_relationship_name="describes",
                target="descriptor",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
    NodeValidation(
        node_type="sample-run-configuration",
        min=0,
        validations=[
            EmbeddedRefValidation(
                node_type="sample-run-configuration",
                node_property_name="protocol_ref",
                required=True,
                target_ref_types=["protocol"],
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="sample-run-configuration",
                relationship_name="described-as",
                reverse_relationship_name="describes",
                target="descriptor",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
    NodeValidation(
        node_type="study",
        min=1,
        validations=[
            EmbeddedRefValidation(
                node_type="study",
                node_property_name="created_by_ref",
                required=False,
                target_ref_types=["data-provider"],
            ),
            NodePropertyValidation(
                node_type="study",
                node_property_name="mhd_identifier",
                contraints=PropertyConstraint(required=True, min_length=8),
            ),
            NodePropertyValidation(
                node_type="study",
                node_property_name="repository_identifier",
                contraints=PropertyConstraint(required=True, min_length=2),
            ),
            EmbeddedRefValidation(
                node_type="study",
                node_property_name="protocol_refs",
                required=False,
                target_ref_types=["protocol"],
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="study",
                relationship_name="provided-by",
                reverse_relationship_name="provides",
                target="data-provider",
                min=0,
                min_for_each_source=1,
                max_for_each_source=1,
            ),
            RelationshipValidation(
                source="study",
                relationship_name="has-assay",
                reverse_relationship_name="part-of",
                target="assay",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="study",
                relationship_name="funded-by",
                reverse_relationship_name="funds",
                target="organization",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="study",
                relationship_name="has-characteristic-definition",
                reverse_relationship_name="used-in",
                target="characteristic-definition",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="study",
                relationship_name="has-derived-data-file",
                reverse_relationship_name="created-in",
                target="derived-data-file",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="study",
                relationship_name="described-as",
                reverse_relationship_name="describes",
                target="descriptor",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="study",
                relationship_name="has-factor-definition",
                reverse_relationship_name="used-in",
                target="factor-definition",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="study",
                relationship_name="has-repository-keyword",
                reverse_relationship_name="keyword-of",
                target="descriptor",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="study",
                relationship_name="has-submitter-keyword",
                reverse_relationship_name="keyword-of",
                target="descriptor",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="study",
                relationship_name="reports",
                reverse_relationship_name="reported-in",
                target="metabolite",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="study",
                relationship_name="has-metadata-file",
                reverse_relationship_name="describes",
                target="metadata-file",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="study",
                relationship_name="has-contributor",
                reverse_relationship_name="contributes",
                target="person",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="study",
                relationship_name="has-principal-investigator",
                reverse_relationship_name="principal-investigator-of",
                target="person",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="study",
                relationship_name="submitted-by",
                reverse_relationship_name="submits",
                target="person",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="study",
                relationship_name="part-of",
                reverse_relationship_name="has-study",
                target="project",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="study",
                relationship_name="has-protocol",
                reverse_relationship_name="used-in",
                target="protocol",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="study",
                relationship_name="has-publication",
                reverse_relationship_name="describes",
                target="publication",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="study",
                relationship_name="has-raw-data-file",
                reverse_relationship_name="created-in",
                target="raw-data-file",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="study",
                relationship_name="has-result-file",
                reverse_relationship_name="created-in",
                target="result-file",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="study",
                relationship_name="has-sample",
                reverse_relationship_name="used-in",
                target="sample",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="study",
                relationship_name="has-supplementary-file",
                reverse_relationship_name="created-in",
                target="supplementary-file",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
    NodeValidation(
        node_type="specimen",
        min=0,
        validations=[
            NodePropertyValidation(
                node_type="specimen",
                node_property_name="name",
                contraints=PropertyConstraint(required=True, min_length=1),
            ),
            NodePropertyValidation(
                node_type="specimen",
                node_property_name="repository_identifier",
                contraints=PropertyConstraint(required=True, min_length=1),
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="specimen",
                relationship_name="described-as",
                reverse_relationship_name="describes",
                target="descriptor",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="specimen",
                relationship_name="has-characteristic-value",
                reverse_relationship_name="value-of",
                target="characteristic-value",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="specimen",
                relationship_name="source-of",
                reverse_relationship_name="derived-from",
                target="sample",
                min=0,
                min_for_each_source=1,
            ),
            RelationshipValidation(
                source="specimen",
                relationship_name="derived-from",
                reverse_relationship_name="source-of",
                target="subject",
                min=0,
                min_for_each_source=1,
            ),
        ],
    ),
    NodeValidation(
        node_type="subject",
        min=0,
        validations=[
            NodePropertyValidation(
                node_type="subject",
                node_property_name="name",
                contraints=PropertyConstraint(required=True, min_length=1),
            ),
            NodePropertyValidation(
                node_type="subject",
                node_property_name="repository_identifier",
                contraints=PropertyConstraint(required=True, min_length=1),
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="subject",
                relationship_name="described-as",
                reverse_relationship_name="describes",
                target="descriptor",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="subject",
                relationship_name="has-characteristic-value",
                reverse_relationship_name="value-of",
                target="characteristic-value",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="subject",
                relationship_name="source-of",
                reverse_relationship_name="derived-from",
                target="sample",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="subject",
                relationship_name="has-factor-value",
                reverse_relationship_name="value-of",
                target="factor-value",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="subject",
                relationship_name="source-of",
                reverse_relationship_name="derived-from",
                target="specimen",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
    NodeValidation(
        node_type="supplementary-file",
        min=0,
        validations=[
            NodePropertyValidation(
                node_type="supplementary-file",
                node_property_name="url_list",
                contraints=PropertyConstraint(required=True, min_length=1),
            ),
            NodePropertyValidation(
                node_type="supplementary-file",
                node_property_name="name",
                contraints=PropertyConstraint(required=True, min_length=2),
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="supplementary-file",
                relationship_name="described-as",
                reverse_relationship_name="describes",
                target="descriptor",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="supplementary-file",
                relationship_name="created-in",
                reverse_relationship_name="has-supplementary-file",
                target="study",
                min=0,
                min_for_each_source=1,
            ),
            RelationshipValidation(
                source="supplementary-file",
                relationship_name="referenced-in",
                reverse_relationship_name="references",
                target="metadata-file",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
]

MHD_BASE_VALIDATION_V0_1.cv_nodes = [
    CvNodeValidation(
        node_type="characteristic-type",
        min=0,
        validations=[
            NodePropertyValidation(
                node_type="characteristic-type",
                node_property_name="name",
                contraints=PropertyConstraint(required=True),
            ),
            CvTermValidation(
                node_type="characteristic-type",
                validation=AllowedCvTerms(
                    cv_terms=list(COMMON_CHARACTERISTIC_DEFINITIONS.values())
                ),
                condition=[
                    FilterCondition(
                        name="Characteristic Type",
                        relationship_name="[embedded].characteristic_type_ref",
                        start_node_type="characteristic-definition",
                        expression="characteristic_type_ref",
                    )
                ],
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="characteristic-type",
                relationship_name="type-of",
                reverse_relationship_name="has-type",
                target="characteristic-definition",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
    CvNodeValidation(
        node_type="characteristic-value",
        has_value=True,
        min=0,
        validations=[
            CvTermValidation(
                node_type="characteristic-value",
                validation=AllowAnyCvTerm(
                    allowed_placeholder_values=[CvTermPlaceholder()],
                    allowed_other_sources=["wikidata", "ILX"],
                ),
            )
        ],
        relationships=[
            RelationshipValidation(
                source="characteristic-value",
                relationship_name="instance-of",
                reverse_relationship_name="has-instance",
                target="characteristic-definition",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="characteristic-value",
                relationship_name="value-of",
                reverse_relationship_name="has-characteristic-value",
                target="subject",
                min=0,
                min_for_each_source=1,
            ),
            RelationshipValidation(
                source="characteristic-value",
                relationship_name="value-of",
                reverse_relationship_name="has-characteristic-value",
                target="specimen",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="characteristic-value",
                relationship_name="value-of",
                reverse_relationship_name="has-characteristic-value",
                target="sample",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
    CvNodeValidation(
        node_type="data-provider",
        has_value=True,
        value_required=True,
        min=1,
        validations=[
            NodePropertyValidation(
                node_type="data-provider",
                node_property_name="value",
                contraints=PropertyConstraint(required=True, allowed_types="str"),
            ),
            CvTermValidation(
                node_type="data-provider",
                validation=AllowAnyCvTerm(allowed_other_sources=["wikidata", "ILX"]),
            ),
            CvTermValidation(
                node_type="data-provider",
                validation=AllowAnyCvTerm(allowed_other_sources=["wikidata", "ILX"]),
                condition=[
                    FilterCondition(
                        name="Data Provider",
                        relationship_name="[embedded].created_by_ref",
                        start_node_type=None,
                        expression="created_by_ref",
                    )
                ],
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="data-provider",
                relationship_name="provides",
                reverse_relationship_name="provided-by",
                target="study",
                min=0,
                min_for_each_source=1,
                max_for_each_source=1,
            ),
        ],
    ),
    CvNodeValidation(
        node_type="descriptor",
        min=0,
        validations=[
            CvTermValidation(
                node_type="descriptor",
                validation=AllowedChildrenCvTerms(
                    parent_cv_terms=[
                        ParentCvTerm(
                            cv_term=CvTerm(
                                source="EDAM",
                                accession="EDAM:1915",
                                name="Format",
                            ),
                            index_cv_terms=False,
                        )
                    ]
                ),
                condition=[
                    FilterCondition(
                        name="File Format",
                        relationship_name="[embedded].format_ref",
                        start_node_type=None,
                        expression="format_ref",
                    )
                ],
            ),
            CvTermValidation(
                node_type="descriptor",
                validation=AllowedChildrenCvTerms(
                    parent_cv_terms=[
                        ParentCvTerm(
                            cv_term=CvTerm(
                                source="EDAM",
                                accession="EDAM:1915",
                                name="Format",
                            ),
                            index_cv_terms=False,
                        )
                    ]
                ),
                condition=[
                    FilterCondition(
                        name="File Format",
                        relationship_name="[embedded].compression_format_refs",
                        start_node_type=None,
                        expression="compression_format_refs",
                    )
                ],
            ),
            CvTermValidation(
                node_type="descriptor",
                validation=AllowedCvTerms(cv_terms=list(COMMON_ASSAY_TYPES.values())),
                condition=[
                    FilterCondition(
                        name="Analysis Type",
                        relationship_name="[embedded].assay_type_ref",
                        start_node_type="assay",
                        expression="assay_type_ref",
                    )
                ],
            ),
            CvTermValidation(
                node_type="descriptor",
                validation=AllowedCvTerms(
                    cv_terms=list(COMMON_MEASUREMENT_TYPES.values())
                ),
                condition=[
                    FilterCondition(
                        name="Measurement Type",
                        relationship_name="[embedded].measurement_type_ref",
                        start_node_type="assay",
                        expression="measurement_type_ref",
                    )
                ],
            ),
            CvTermValidation(
                node_type="descriptor",
                validation=AllowedCvTerms(
                    cv_terms=list(COMMON_TECHNOLOGY_TYPES.values())
                ),
                condition=[
                    FilterCondition(
                        name="Technology Type",
                        relationship_name="[embedded].technology_type_ref",
                        start_node_type="assay",
                        expression="technology_type_ref",
                    )
                ],
            ),
            CvTermValidation(
                node_type="descriptor",
                validation=AllowedCvTerms(cv_terms=list(COMMON_OMICS_TYPES.values())),
                condition=[
                    FilterCondition(
                        name="Omics Type",
                        relationship_name="[embedded].omics_type_ref",
                        start_node_type="assay",
                        expression="omics_type_ref",
                    )
                ],
            ),
            CvTermValidation(
                node_type="descriptor",
                validation=AllowAnyCvTerm(
                    allowed_other_sources=["wikidata", "ILX"],
                    allowed_placeholder_values=[CvTermPlaceholder()],
                ),
                condition=[
                    FilterCondition(
                        name="Keyword",
                        relationship_name="has-submitter-keyword",
                        start_node_type="study",
                    )
                ],
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="descriptor",
                relationship_name="describes",
                reverse_relationship_name="described-as",
                target="assay",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="descriptor",
                relationship_name="describes",
                reverse_relationship_name="described-as",
                target="study",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="descriptor",
                relationship_name="describes",
                reverse_relationship_name="described-as",
                target="metadata-file",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="descriptor",
                relationship_name="describes",
                reverse_relationship_name="described-as",
                target="raw-data-file",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="descriptor",
                relationship_name="describes",
                reverse_relationship_name="described-as",
                target="derived-data-file",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="descriptor",
                relationship_name="describes",
                reverse_relationship_name="described-as",
                target="supplementary-file",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="descriptor",
                relationship_name="describes",
                reverse_relationship_name="described-as",
                target="result-file",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="descriptor",
                relationship_name="describes",
                reverse_relationship_name="described-as",
                target="metabolite",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="descriptor",
                relationship_name="describes",
                reverse_relationship_name="described-as",
                target="organization",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="descriptor",
                relationship_name="describes",
                reverse_relationship_name="described-as",
                target="person",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="descriptor",
                relationship_name="describes",
                reverse_relationship_name="described-as",
                target="project",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="descriptor",
                relationship_name="describes",
                reverse_relationship_name="described-as",
                target="publication",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="descriptor",
                relationship_name="describes",
                reverse_relationship_name="described-as",
                target="protocol",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="descriptor",
                relationship_name="describes",
                reverse_relationship_name="described-as",
                target="sample",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="descriptor",
                relationship_name="describes",
                reverse_relationship_name="described-as",
                target="subject",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="descriptor",
                relationship_name="describes",
                reverse_relationship_name="described-as",
                target="sample-run",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="descriptor",
                relationship_name="describes",
                reverse_relationship_name="described-as",
                target="sample-run-configuration",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="descriptor",
                relationship_name="describes",
                reverse_relationship_name="described-as",
                target="metabolite",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="descriptor",
                relationship_name="keyword-of",
                reverse_relationship_name="has-submitter-keyword",
                target="study",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="descriptor",
                relationship_name="keyword-of",
                reverse_relationship_name="has-repository-keyword",
                target="study",
                min=0,
                min_for_each_source=0,
            ),
            RelationshipValidation(
                source="descriptor",
                relationship_name="keyword-of",
                reverse_relationship_name="has-repository-keyword",
                target="specimen",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
    CvNodeValidation(
        node_type="factor-type",
        min=0,
        validations=[
            NodePropertyValidation(
                node_type="factor-type",
                node_property_name="name",
                contraints=PropertyConstraint(required=True),
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="factor-type",
                relationship_name="type-of",
                reverse_relationship_name="has-type",
                target="factor-definition",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
    CvNodeValidation(
        node_type="factor-value",
        has_value=True,
        min=0,
        validations=[
            CvTermValidation(
                node_type="factor-value",
                validation=AllowAnyCvTerm(
                    allowed_other_sources=["wikidata", "ILX"],
                    allowed_placeholder_values=[CvTermPlaceholder()],
                ),
            ),
            CvTermValidation(
                node_type="factor-value",
                validation=AllowedCvList(
                    source_names=["DOID", "HP", "MP", "SNOMED"],
                    allowed_other_sources=["wikidata", "ILX"],
                ),
                condition=[
                    FilterCondition(
                        name="Disease",
                        relationship_name="has-instance",
                        start_node_type="factor-definition",
                        expression="factor_type_ref.accession",
                        expression_value="EFO:0000408",
                    )
                ],
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="factor-value",
                relationship_name="instance-of",
                reverse_relationship_name="has-instance",
                target="factor-definition",
                min=0,
                min_for_each_source=1,
            ),
            RelationshipValidation(
                source="factor-value",
                relationship_name="value-of",
                reverse_relationship_name="has-factor-value",
                target="sample",
                min=0,
                min_for_each_source=1,
            ),
            RelationshipValidation(
                source="factor-value",
                relationship_name="value-of",
                reverse_relationship_name="has-factor-value",
                target="specimen",
                min=0,
                min_for_each_source=1,
            ),
            RelationshipValidation(
                source="factor-value",
                relationship_name="value-of",
                reverse_relationship_name="has-factor-value",
                target="subject",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
    CvNodeValidation(
        node_type="metabolite-identifier",
        min=0,
        validations=[
            CvTermValidation(
                node_type="metabolite-identifier",
                validation=AllowedChildrenCvTerms(
                    parent_cv_terms=[
                        ParentCvTerm(
                            cv_term=CvTerm(
                                source="CHEMINF",
                                accession="CHEMINF:000464",
                                name="chemical database identifier",
                            ),
                            index_cv_terms=False,
                        )
                    ]
                ),
                condition=[
                    FilterCondition(
                        name="Reported Metabolite Identifier",
                        relationship_name="identified-as",
                        start_node_type="metabolite",
                    )
                ],
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="metabolite-identifier",
                relationship_name="reported-identifier-of",
                reverse_relationship_name="identified-as",
                target="metabolite",
                min=0,
                min_for_each_source=1,
            ),
        ],
    ),
    CvNodeValidation(
        node_type="parameter-type",
        min=0,
        validations=[
            NodePropertyValidation(
                node_type="parameter-type",
                node_property_name="name",
                contraints=PropertyConstraint(required=True),
            ),
            CvTermValidation(
                node_type="parameter-type",
                validation=AllowedCvTerms(
                    cv_terms=list(COMMON_PARAMETER_DEFINITIONS.values()),
                    allowed_other_sources=["wikidata", "ILX"],
                ),
            ),
            CvTermValidation(
                node_type="parameter-type",
                validation=AllowedCvTerms(
                    cv_terms=list(COMMON_PROTOCOL_PARAMETERS["CHMO:0000470"].values())
                ),
                condition=[
                    FilterCondition(
                        name="Mass spectrometry protocol",
                        relationship_name="[embedded].parameter_type_ref",
                        start_node_type="parameter-definition",
                        expression="[used-in].protocol_type_ref.accession",
                        expression_value="CHMO:0000470",
                    ),
                ],
            ),
            CvTermValidation(
                node_type="parameter-type",
                validation=AllowedCvTerms(
                    cv_terms=list(COMMON_PROTOCOL_PARAMETERS["CHMO:0001000"].values())
                ),
                condition=[
                    FilterCondition(
                        name="Chromatography protocol",
                        relationship_name="[embedded].parameter_definition_refs.parameter_type_ref",
                        start_node_type="protocol",
                        expression="protocol_type_ref.accession",
                        expression_value="CHMO:0001000",
                    )
                ],
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="parameter-type",
                relationship_name="type-of",
                reverse_relationship_name="has-type",
                target="parameter-definition",
                min=0,
                min_for_each_source=0,
            ),
        ],
    ),
    CvNodeValidation(
        node_type="parameter-value",
        has_value=True,
        min=0,
        validations=[
            CvTermValidation(
                node_type="parameter-value",
                validation=AllowAnyCvTerm(
                    allowed_placeholder_values=[CvTermPlaceholder()],
                    allowed_other_sources=["wikidata", "ILX"],
                ),
            ),
            CvTermValidation(
                node_type="parameter-value",
                validation=AllowedChildrenCvTerms(
                    parent_cv_terms=[
                        ParentCvTerm(
                            cv_term=CvTerm(
                                source="MS",
                                accession="MS:1000031",
                                name="instrument model",
                            ),
                            excluded_cv_terms=[
                                CvTerm(
                                    source="MS",
                                    accession="MS:1000491",
                                    name="Dionex instrument model",
                                ),
                                CvTerm(
                                    source="MS",
                                    accession="MS:1000488",
                                    name="Hitachi instrument model",
                                ),
                            ],
                            allow_only_leaf=True,
                        ),
                    ]
                ),
                condition=[
                    FilterCondition(
                        name="Mass spectrometry instrument",
                        relationship_name="has-instance",
                        start_node_type="parameter-definition",
                        expression="parameter_type_ref.accession",
                        expression_value="MSIO:0000171",
                    )
                ],
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="parameter-value",
                relationship_name="instance-of",
                reverse_relationship_name="has-instance",
                target="parameter-definition",
                min=0,
                min_for_each_source=1,
            ),
        ],
    ),
    CvNodeValidation(
        node_type="protocol-type",
        min=0,
        validations=[
            NodePropertyValidation(
                node_type="protocol-type",
                node_property_name="name",
                contraints=PropertyConstraint(required=True),
            ),
            CvTermValidation(
                node_type="protocol-type",
                validation=AllowedCvTerms(
                    cv_terms=list(COMMON_PROTOCOLS.values()),
                    allowed_other_sources=["wikidata", "ILX"],
                ),
                condition=[
                    FilterCondition(
                        name="Protocol Type",
                        relationship_name="[embedded].protocol_type_ref",
                        start_node_type="protocol",
                        expression="protocol_type_ref",
                    )
                ],
            ),
        ],
        relationships=[
            RelationshipValidation(
                source="protocol-type",
                relationship_name="type-of",
                reverse_relationship_name="has-type",
                target="protocol",
                min=0,
                min_for_each_source=1,
            ),
        ],
    ),
]

MHD_BASE_VALIDATION_V0_1.mhd_nodes.sort(key=lambda x: x.node_type)

MHD_BASE_VALIDATION_V0_1.cv_nodes.sort(key=lambda x: x.node_type)



import logging
from pathlib import Path
from typing import Any, OrderedDict

from pydantic import AnyUrl, BaseModel

from mhd_model.model.definitions import (
    ANNOUNCEMENT_FILE_V0_1_DEFAULT_SCHEMA_NAME,
    ANNOUNCEMENT_FILE_V0_1_LEGACY_PROFILE_NAME,
)
from mhd_model.model.v0_1.announcement.profiles.base.fields import (
    ExtendedCvTermKeyValue,
)
from mhd_model.model.v0_1.announcement.profiles.base.profile import (
    AnnouncementBaseFile,
    AnnouncementDerivedDataFile,
    AnnouncementMetadataFile,
    AnnouncementProtocol,
    AnnouncementPublication,
    AnnouncementRawDataFile,
    AnnouncementReportedMetabolite,
    AnnouncementResultFile,
    AnnouncementSupplementaryFile,
)
from mhd_model.model.v0_1.announcement.profiles.legacy.profile import (
    AnnouncementContact,
    AnnouncementLegacyProfile,
)
from mhd_model.model.v0_1.dataset.profiles.base import graph_nodes
from mhd_model.model.v0_1.dataset.profiles.base.base import (
    BaseMhdModel,
    BaseMhdRelationship,
    CvTerm,
    CvTermValue,
    IdentifiableMhdModel,
)
from mhd_model.model.v0_1.dataset.profiles.legacy.profile import MhDatasetLegacyProfile
from mhd_model.model.v0_1.rules.cv_definitions import (
    CONTROLLED_CV_DEFINITIONS,
    OTHER_CONTROLLED_CV_DEFINITIONS,
)
from mhd_model.model.v0_1.rules.managed_cv_terms import (
    COMMON_ASSAY_TYPES,
    COMMON_MEASUREMENT_TYPES,
    COMMON_OMICS_TYPES,
    COMMON_TECHNOLOGY_TYPES,
    MISSING_PUBLICATION_REASON,
)
from mhd_model.shared.model import CvDefinition

logger = logging.getLogger(__name__)


def get_characteristic_values(
    all_nodes_map: dict[str, IdentifiableMhdModel],
    relationships_map: dict[str, BaseMhdRelationship],
) -> list[ExtendedCvTermKeyValue]:
    study_characteristics = set()
    characteristic_values: OrderedDict[str, list[str]] = OrderedDict()
    for rel in relationships_map.values():
        source = all_nodes_map[rel.source_ref]
        if rel.relationship_name == "has-characteristic-definition":
            if isinstance(source, graph_nodes.Study):
                study_characteristics.add(rel.target_ref)

        if rel.relationship_name == "has-instance":
            if isinstance(source, graph_nodes.CharacteristicDefinition):
                if rel.source_ref not in characteristic_values:
                    characteristic_values[rel.source_ref] = []
                characteristic_values[rel.source_ref].append(rel.target_ref)
    referenced_characteristics = {
        x: y
        for x, y in characteristic_values.items()
        if x in study_characteristics and y
    }
    characteristic_keys = [
        (x, all_nodes_map[x]) for x, y in referenced_characteristics.items()
    ]
    characteristic_keys.sort(key=lambda x: x[1].name)
    announcement_characteristics = []
    for char_id, characteristic in characteristic_keys:
        value_ids = referenced_characteristics[char_id]
        values = []
        for x in value_ids:
            val = all_nodes_map[x]
            if val.id_.startswith("cv-value--"):
                val = CvTermValue.model_validate(val.model_dump(by_alias=True))
            else:
                val = CvTerm.model_validate(val.model_dump(by_alias=True))
            values.append(val)
        type_node = all_nodes_map.get(characteristic.characteristic_type_ref, None)
        key = CvTerm.model_validate(type_node.model_dump(by_alias=True))
        announcement_characteristics.append(
            ExtendedCvTermKeyValue(key=key, values=values)
        )

    return announcement_characteristics


def get_keywords(
    all_nodes_map: dict[str, IdentifiableMhdModel],
    relationship_name_map: dict[str, dict[str, BaseMhdRelationship]],
) -> list[CvTerm]:
    keywords = []
    if "has-submitter-keyword" in relationship_name_map:
        for rel in relationship_name_map.get("has-submitter-keyword").values():
            source = all_nodes_map.get(rel.source_ref)
            if source:
                if isinstance(source, graph_nodes.Study):
                    keyword_node = all_nodes_map.get(rel.target_ref)
                    if keyword_node:
                        keyword = CvTerm.model_validate(
                            keyword_node.model_dump(by_alias=True)
                        )
                        keywords.append(keyword)
    return keywords


def get_study_factors(
    all_nodes_map: dict[str, IdentifiableMhdModel],
    relationships_map: dict[str, BaseMhdRelationship],
):
    study_factors = set()
    factors: OrderedDict[str, list[str]] = OrderedDict()
    for rel in relationships_map.values():
        source = all_nodes_map[rel.source_ref]
        if rel.relationship_name == "has-factor-definition":
            if isinstance(source, graph_nodes.Study):
                study_factors.add(rel.target_ref)

        if rel.relationship_name == "has-instance":
            if isinstance(source, graph_nodes.FactorDefinition):
                if rel.source_ref not in factors:
                    factors[rel.source_ref] = []
                factors[rel.source_ref].append(rel.target_ref)
    referenced_factors = {x: y for x, y in factors.items() if x in study_factors and y}
    factor_keys = [(x, all_nodes_map[x]) for x, y in referenced_factors.items()]
    factor_keys.sort(key=lambda x: x[1].name)
    announcement_factors = []
    for factor_id, factor in factor_keys:
        value_ids = referenced_factors[factor_id]
        values = []
        for x in value_ids:
            val = all_nodes_map[x]
            if isinstance(val, CvTermValue):
                val = CvTermValue.model_validate(val.model_dump(by_alias=True))
            else:
                val = CvTerm.model_validate(val.model_dump(by_alias=True))
            values.append(val)
        type_node = all_nodes_map.get(factor.factor_type_ref, None)
        key = CvTerm.model_validate(type_node.model_dump(by_alias=True))
        announcement_factors.append(ExtendedCvTermKeyValue(key=key, values=values))
    return announcement_factors


def get_protocols(
    all_nodes_map: dict[str, IdentifiableMhdModel],
    relationship_name_map: dict[str, BaseMhdRelationship],
    type_map: dict[str, dict[IdentifiableMhdModel]],
    study: graph_nodes.Study,
):
    protocols: list[AnnouncementProtocol] = []
    if not study.protocol_refs:
        return protocols

    for ref in study.protocol_refs:
        if ref not in type_map["protocol"]:
            logger.error("Protocol %s is not defined in dataset.", ref)
        protocol_parameters = []
        protocol_node: graph_nodes.Protocol = type_map["protocol"].get(ref)
        protocol_type_object = all_nodes_map.get(protocol_node.protocol_type_ref, None)
        # if not protocol_node.parameter_definition_refs:
        #     continue
        protocol_parameters = []
        if protocol_node.parameter_definition_refs:
            for definition_key in protocol_node.parameter_definition_refs:
                if definition_key not in all_nodes_map:
                    continue
                definition = all_nodes_map[definition_key]
                if not isinstance(definition, graph_nodes.ParameterDefinition):
                    continue
                vals = []
                for rel in relationship_name_map["has-instance"].values():
                    if rel.source_ref == definition.id_:
                        val = None
                        val_obj = all_nodes_map.get(rel.target_ref)
                        if isinstance(val_obj, CvTerm):
                            val = CvTerm.model_validate(
                                val_obj.model_dump(by_alias=True)
                            )
                        elif isinstance(val_obj, CvTermValue):
                            val = CvTermValue.model_validate(
                                val_obj.model_dump(by_alias=True)
                            )
                        if not val:
                            continue
                        vals.append(val)
                # if vals:
                def_type = all_nodes_map.get(definition.parameter_type_ref)
                key = CvTerm.model_validate(def_type.model_dump(by_alias=True))
                param = ExtendedCvTermKeyValue(
                    key=key,
                    values=vals or None,
                )
                protocol_parameters.append(param)
        if not protocol_parameters:
            protocol_parameters = None
        else:
            protocol_parameters.sort(key=lambda x: x.key.name)
        protocol_type = CvTerm.model_validate(
            protocol_type_object.model_dump(by_alias=True)
        )
        protocol = AnnouncementProtocol(
            name=protocol_node.name,
            protocol_type=protocol_type,
            description=protocol_node.description,
            protocol_parameters=protocol_parameters,
        )
        protocols.append(protocol)

    return protocols


def convert_file(
    all_nodes_map,
    type_map: dict[str, dict[str, IdentifiableMhdModel]],
    file_type_name: str,
    ref: str,
    file_class: type[AnnouncementBaseFile],
):
    if file_type_name not in type_map or ref not in type_map[file_type_name]:
        return None
    item: graph_nodes.BaseFile = type_map[file_type_name][ref]
    url_list = item.url_list
    format = None
    if item.format_ref in all_nodes_map:
        format_node: BaseMhdModel = all_nodes_map[item.format_ref]
        format = CvTerm.model_validate(format_node.model_dump(by_alias=True))
    compressions = []
    if item.compression_format_refs in all_nodes_map:
        for format_ref in item.compression_format_refs:
            compression_node: BaseMhdModel = all_nodes_map[format_ref]
            compressions.append(
                CvTerm.model_validate(compression_node.model_dump(by_alias=True))
            )

    file = file_class(
        name=item.name,
        url_list=url_list,
        compression_formats=compressions or None,
        format=format,
    )

    return file


def collect_cvterm_sources(obj: BaseModel, cv_sources: set[str]):
    if isinstance(obj, (CvTerm, CvTermValue)):
        source = getattr(obj, "source", None)
        if source:
            cv_sources.add(source)
    elif isinstance(obj, BaseModel):
        for value in obj.__dict__.values():
            collect_cvterm_sources(value, cv_sources)
    elif isinstance(obj, list):
        for item in obj:
            collect_cvterm_sources(item, cv_sources)
    elif isinstance(obj, dict):
        for value in obj.values():
            collect_cvterm_sources(value, cv_sources)


def create_announcement_file(
    mhd_file: dict[str, Any],
    mhd_file_url: str,
    announcement_file_path: str,
    announcement_schema_name: str = ANNOUNCEMENT_FILE_V0_1_DEFAULT_SCHEMA_NAME,
    announcement_profile_uri=ANNOUNCEMENT_FILE_V0_1_LEGACY_PROFILE_NAME,
):
    mhd_dataset = MhDatasetLegacyProfile.model_validate(mhd_file)
    nodes_map: dict[str, IdentifiableMhdModel] = {
        x.id_: x for x in mhd_dataset.graph.nodes
    }
    relationships_map: dict[str, BaseMhdRelationship] = {
        x.id_: x for x in mhd_dataset.graph.relationships
    }

    all_nodes_map: dict[str, BaseMhdModel] = {}
    type_map: dict[str, dict[str, BaseMhdModel]] = {}
    for node in mhd_dataset.graph.nodes:
        if node.type_ not in type_map:
            type_map[node.type_] = {}
        type_map[node.type_][node.id_] = node
        all_nodes_map[node.id_] = node

    relationship_name_map: dict[str, dict[str, BaseMhdRelationship]] = {}
    for rel in mhd_dataset.graph.relationships:
        if rel.relationship_name not in relationship_name_map:
            relationship_name_map[rel.relationship_name] = {}
        relationship_name_map[rel.relationship_name][rel.id_] = rel

    if "study" not in type_map:
        logger.error("Study not found for in the input file")
        return
    study: graph_nodes.Study = list(type_map["study"].values())[0]

    study_assays: list[graph_nodes.Assay] = []
    if "assay" in type_map:
        study_assays = list(type_map["assay"].values())

    publications = get_publications(nodes_map, type_map, relationship_name_map)

    submitters, principal_investigators = get_submitter_and_pi(
        type_map, relationship_name_map
    )
    keywords = get_keywords(all_nodes_map, relationship_name_map)
    assay_types, technology_types, measurement_types, omics_types = (
        get_main_assay_descriptors(nodes_map, study_assays, keywords)
    )

    repository_metadata_file_list = get_file_list(
        all_nodes_map, type_map, "metadata-file", AnnouncementMetadataFile
    )
    result_file_list = get_file_list(
        all_nodes_map, type_map, "result-file", AnnouncementResultFile
    )

    raw_data_file_list = get_file_list(
        all_nodes_map, type_map, "raw-data-file", AnnouncementRawDataFile
    )
    derived_data_file_list = get_file_list(
        all_nodes_map, type_map, "derived-data-file", AnnouncementDerivedDataFile
    )
    supplementary_file_list = get_file_list(
        all_nodes_map,
        type_map,
        "supplementary-file",
        AnnouncementSupplementaryFile,
    )
    reported_metabolites = get_metabolites(type_map, relationship_name_map)
    dataset_url_list = study.dataset_url_list

    protocols = get_protocols(all_nodes_map, relationship_name_map, type_map, study)

    study_factors = get_study_factors(all_nodes_map, relationships_map)
    characteristic_values = get_characteristic_values(all_nodes_map, relationships_map)

    if not mhd_file_url:
        ftp = [str(x) for x in study.dataset_url_list if str(x).startswith("ftp://")]
        if ftp:
            mhd_file_url = f"{ftp[0]}/{study.repository_identifier}.mhd.json"

    announcement = AnnouncementLegacyProfile(
        mhd_identifier=study.mhd_identifier,
        repository_identifier=study.repository_identifier,
        schema_name=announcement_schema_name,
        profile_uri=announcement_profile_uri,
        mhd_metadata_file_url=AnyUrl(mhd_file_url),
        dataset_url_list=dataset_url_list,
        license=study.license,
        title=study.title,
        description=study.description,
        submission_date=study.submission_date,
        public_release_date=study.public_release_date,
        submitters=submitters or None,
        principal_investigators=principal_investigators or None,
        measurement_type=measurement_types or None,
        technology_type=technology_types or None,
        assay_type=assay_types or None,
        omics_type=omics_types or None,
        repository_metadata_file_list=repository_metadata_file_list or None,
        result_file_list=result_file_list or None,
        raw_data_file_list=raw_data_file_list or None,
        derived_data_file_list=derived_data_file_list or None,
        supplementary_file_list=supplementary_file_list or None,
        publications=publications or None,
        submitter_keywords=keywords or None,
        reported_metabolites=reported_metabolites or None,
        protocols=protocols,
        study_factors=study_factors or None,
        characteristic_values=characteristic_values,
    )
    cv_sources = set()
    collect_cvterm_sources(announcement, cv_sources)
    cv_sources = list(cv_sources)
    cv_sources.sort()
    for source in cv_sources:
        if source in CONTROLLED_CV_DEFINITIONS:
            announcement.cv_definitions.append(CONTROLLED_CV_DEFINITIONS[source])
        elif source in OTHER_CONTROLLED_CV_DEFINITIONS:
            announcement.cv_definitions.append(OTHER_CONTROLLED_CV_DEFINITIONS[source])
        else:
            announcement.cv_definitions.append(
                CvDefinition(label=source, alternative_labels=[source.lower()])
            )
    logger.info("Writing to %s", announcement_file_path)
    Path(announcement_file_path).parent.mkdir(parents=True, exist_ok=True)
    with Path(announcement_file_path).open("w") as f:
        f.write(
            announcement.model_dump_json(indent=2, by_alias=True, exclude_none=True)
        )


def get_metabolites(
    type_map, relationship_name_map
) -> None | list[AnnouncementReportedMetabolite]:
    identification_map = {}
    identification_links = relationship_name_map.get("identified-as")
    items = type_map.get("metabolite-identification")
    if identification_links and items:
        for ref in identification_links:
            item = identification_links[ref]
            if item.target_ref in items:
                identification = items[item.target_ref]
                if item.source_ref not in identification_map:
                    identification_map[item.source_ref] = []
                identification_map[item.source_ref].append(identification)
    reported_metabolites: list[AnnouncementReportedMetabolite] = []
    if "metabolite" in type_map:
        for ref in type_map["metabolite"]:
            met = type_map["metabolite"][ref]
            item = AnnouncementReportedMetabolite(name=met.name)
            reported_metabolites.append(item)

            if ref in identification_map:
                identifications = identification_map[ref]
                item.database_identifiers = [
                    CvTermValue.model_validate(x.model_dump(by_alias=True))
                    for x in identifications
                ]

        if reported_metabolites:
            reported_metabolites.sort(key=lambda x: x.name)
    return reported_metabolites or None


def get_file_list(
    all_nodes_map: dict[str, IdentifiableMhdModel],
    type_map: dict[str, dict[str, BaseMhdModel]],
    node_type: str,
    file_class: type[AnnouncementBaseFile],
) -> list[AnnouncementBaseFile]:
    file_list = []
    if node_type in type_map:
        for ref in type_map[node_type]:
            file_object = convert_file(
                all_nodes_map, type_map, node_type, ref, file_class
            )
            if file_object:
                file_list.append(file_object)
    return file_list


def get_main_assay_descriptors(
    nodes_map: dict[str, IdentifiableMhdModel],
    study_assays: list[graph_nodes.Assay],
    keywords: list[CvTerm],
) -> tuple[list[CvTerm], list[CvTerm], list[CvTerm], list[CvTerm]]:
    assay_types: dict[str, CvTerm] = {}
    technology_types: dict[str, CvTerm] = {}
    measurement_types: dict[str, CvTerm] = {}
    omics_types: dict[str, CvTerm] = {}
    for item in study_assays:
        if item.assay_type_ref in nodes_map:
            assay_type: graph_nodes.CvTermObject = nodes_map[item.assay_type_ref]
            if assay_type.accession not in assay_types:
                term = COMMON_ASSAY_TYPES.get(assay_type.accession, None)
                if not term:
                    term = CvTerm.model_validate(assay_type.model_dump(by_alias=True))
                assay_types[term.accession] = term

        if item.technology_type_ref in nodes_map:
            technology_type: graph_nodes.CvTermObject = nodes_map[
                item.technology_type_ref
            ]
            if technology_type.accession not in technology_types:
                term = COMMON_TECHNOLOGY_TYPES.get(technology_type.accession)
                if not term:
                    term = CvTerm.model_validate(
                        technology_type.model_dump(by_alias=True)
                    )
                technology_types[term.accession] = term
        if item.measurement_type_ref in nodes_map:
            measurement_type: graph_nodes.CvTermObject = nodes_map[
                item.technology_type_ref
            ]
            if measurement_type.accession not in measurement_types:
                term = COMMON_MEASUREMENT_TYPES.get(measurement_type.accession)
                if not term:
                    term = CvTerm.model_validate(
                        technology_type.model_dump(by_alias=True)
                    )
                measurement_types[term.accession] = term

    for keyword in keywords:
        if "untargetted" in keyword.name.lower():
            measurement_types["MSIO:0000101"] = COMMON_MEASUREMENT_TYPES["MSIO:0000101"]
        elif "targetted" in keyword.name.lower():
            measurement_types["MSIO:0000100"] = COMMON_MEASUREMENT_TYPES["MSIO:0000101"]
        for accession, value in COMMON_OMICS_TYPES.items():
            if accession.lower() == keyword.name.lower():
                omics_types[accession] = value

    return (
        list(assay_types.values()),
        list(technology_types.values()),
        list(measurement_types.values()),
        list(omics_types.values()),
    )


def get_submitter_and_pi(
    type_map: dict[str, dict[str, BaseMhdModel]],
    relationship_name_map: dict[str, dict[str, BaseMhdRelationship]],
):
    submitter_links: list[BaseMhdRelationship] = []
    if "submits" in relationship_name_map:
        submitter_links = list(relationship_name_map["submits"].values())

    submitters = []

    principal_investigators = []
    if "person" in type_map:
        for item in submitter_links:
            if item.source_ref in type_map["person"]:
                submitter = type_map["person"][item.source_ref]
                submitters.append(
                    AnnouncementContact.model_validate(submitter, from_attributes=True)
                )

        if "principal-investigator-of" in relationship_name_map:
            pi_links: list[BaseMhdRelationship] = list(
                relationship_name_map["principal-investigator-of"].values()
            )
            for item in pi_links:
                if item.source_ref in type_map["person"]:
                    pi = type_map["person"][item.source_ref]
                    principal_investigators.append(
                        AnnouncementContact.model_validate(pi, from_attributes=True)
                    )

    return submitters, principal_investigators


def get_publications(
    nodes_map: dict[str, IdentifiableMhdModel],
    type_map: dict[str, dict[str, BaseMhdModel]],
    relationship_name_map: dict[str, dict[str, BaseMhdRelationship]],
) -> CvTerm | list[AnnouncementPublication]:
    publications: list[AnnouncementPublication] = []
    if "publication" in type_map:
        graph_publications: list[graph_nodes.Publication] = list(
            type_map["publication"].values()
        )
        for node in graph_publications:
            item = AnnouncementPublication.model_validate(
                node.model_dump(by_alias=True)
            )
            publications.append(item)
    if publications:
        return publications
    publication_status = None
    if "defined-as" in relationship_name_map:
        definitions = list(relationship_name_map["defined-as"].values())
        if definitions:
            for definition in definitions:
                source_node = nodes_map[definition.source_ref]
                target_node = nodes_map[definition.target_ref]
                if (
                    source_node.type_ == "study"
                    and target_node.type_ == "descriptor"
                    and target_node.accession in MISSING_PUBLICATION_REASON
                ):
                    publication_status = CvTerm.model_validate(
                        target_node.model_dump(by_alias=True)
                    )
                    break
    if not publication_status:
        logger.warning("Publication status is not found in study")
    return publication_status



import logging
from pathlib import Path
from typing import Any, OrderedDict

from pydantic import BaseModel

from mhd_model.model.definitions import (
    ANNOUNCEMENT_FILE_V0_1_DEFAULT_SCHEMA_NAME,
    ANNOUNCEMENT_FILE_V0_1_LEGACY_PROFILE_NAME,
)
from mhd_model.model.v0_1.announcement.profiles.base.fields import (
    ExtendedCvTermKeyValue,
)
from mhd_model.model.v0_1.announcement.profiles.base.profile import (
    AnnouncementBaseFile,
    AnnouncementBaseProfile,
    AnnouncementContact,
    AnnouncementDerivedDataFile,
    AnnouncementMetadataFile,
    AnnouncementProtocol,
    AnnouncementPublication,
    AnnouncementRawDataFile,
    AnnouncementReportedMetabolite,
    AnnouncementResultFile,
    AnnouncementSupplementaryFile,
)
from mhd_model.model.v0_1.announcement.profiles.ms.profile import AnnouncementMsProfile
from mhd_model.model.v0_1.dataset.profiles.base import graph_nodes
from mhd_model.model.v0_1.dataset.profiles.base.base import (
    BaseMhdModel,
    BaseMhdRelationship,
    CvTerm,
    CvTermValue,
    IdentifiableMhdModel,
)
from mhd_model.model.v0_1.dataset.profiles.ms.profile import MhDatasetMsProfile
from mhd_model.model.v0_1.rules.cv_definitions import (
    CONTROLLED_CV_DEFINITIONS,
    OTHER_CONTROLLED_CV_DEFINITIONS,
)
from mhd_model.shared.model import CvDefinition

logger = logging.getLogger(__name__)


def update_characteristic_values(
    all_nodes_map: dict[str, IdentifiableMhdModel],
    relationships_map: dict[str, BaseMhdRelationship],
    announcement: AnnouncementMsProfile,
):
    study_characteristics = set()
    characteristic_values: OrderedDict[str, list[str]] = OrderedDict()
    for rel in relationships_map.values():
        source = all_nodes_map[rel.source_ref]
        if rel.relationship_name == "has-characteristic-definition":
            if isinstance(source, graph_nodes.Study):
                study_characteristics.add(rel.target_ref)

        if rel.relationship_name == "has-instance":
            if isinstance(source, graph_nodes.CharacteristicDefinition):
                if rel.source_ref not in characteristic_values:
                    characteristic_values[rel.source_ref] = []
                characteristic_values[rel.source_ref].append(rel.target_ref)
    referenced_characteristics = {
        x: y
        for x, y in characteristic_values.items()
        if x in study_characteristics and y
    }
    characteristic_keys = [
        (x, all_nodes_map[x]) for x, y in referenced_characteristics.items()
    ]
    characteristic_keys.sort(key=lambda x: x[1].name)

    for char_id, characteristic in characteristic_keys:
        value_ids = referenced_characteristics[char_id]
        values = []
        for x in value_ids:
            val = all_nodes_map[x]
            if val.id_.startswith("cv-value--"):
                val = CvTermValue.model_validate(val.model_dump(by_alias=True))
            else:
                val = CvTerm.model_validate(val.model_dump(by_alias=True))
            values.append(val)
        type_node = all_nodes_map.get(characteristic.characteristic_type_ref, None)
        key = CvTerm.model_validate(type_node.model_dump(by_alias=True))
        if not announcement.characteristic_values:
            announcement.characteristic_values = []
        announcement.characteristic_values.append(
            ExtendedCvTermKeyValue(key=key, values=values)
        )


def update_keywords(
    all_nodes_map: dict[str, IdentifiableMhdModel],
    relationship_name_map: dict[str, dict[str, BaseMhdRelationship]],
    announcement: AnnouncementMsProfile,
):
    if "has-submitter-keyword" in relationship_name_map:
        for rel in relationship_name_map.get("has-submitter-keyword").values():
            source = all_nodes_map.get(rel.source_ref)
            if source:
                if isinstance(source, graph_nodes.Study):
                    keyword_node = all_nodes_map.get(rel.target_ref)
                    if keyword_node:
                        keyword = CvTerm.model_validate(
                            keyword_node.model_dump(by_alias=True)
                        )
                        if announcement.submitter_keywords is None:
                            announcement.submitter_keywords = []
                        announcement.submitter_keywords.append(keyword)


def update_study_factors(
    all_nodes_map: dict[str, IdentifiableMhdModel],
    relationships_map: dict[str, BaseMhdRelationship],
    announcement: AnnouncementMsProfile,
):
    study_factors = set()
    factors: OrderedDict[str, list[str]] = OrderedDict()
    for rel in relationships_map.values():
        source = all_nodes_map[rel.source_ref]
        if rel.relationship_name == "has-factor-definition":
            if isinstance(source, graph_nodes.Study):
                study_factors.add(rel.target_ref)

        if rel.relationship_name == "has-instance":
            if isinstance(source, graph_nodes.FactorDefinition):
                if rel.source_ref not in factors:
                    factors[rel.source_ref] = []
                factors[rel.source_ref].append(rel.target_ref)
    referenced_factors = {x: y for x, y in factors.items() if x in study_factors and y}
    factor_keys = [(x, all_nodes_map[x]) for x, y in referenced_factors.items()]
    factor_keys.sort(key=lambda x: x[1].name)

    for factor_id, factor in factor_keys:
        value_ids = referenced_factors[factor_id]
        values = []
        for x in value_ids:
            val = all_nodes_map[x]
            if isinstance(val, CvTermValue):
                val = CvTermValue.model_validate(val.model_dump(by_alias=True))
            else:
                val = CvTerm.model_validate(val.model_dump(by_alias=True))
            values.append(val)
        type_node = all_nodes_map.get(factor.factor_type_ref, None)
        key = CvTerm.model_validate(type_node.model_dump(by_alias=True))
        if not announcement.study_factors:
            announcement.study_factors = []
        announcement.study_factors.append(
            ExtendedCvTermKeyValue(key=key, values=values)
        )


def update_protocol_parameters(
    all_nodes_map: dict[str, IdentifiableMhdModel],
    relationship_name_map: dict[str, BaseMhdRelationship],
    type_map: dict[str, dict[IdentifiableMhdModel]],
    study: graph_nodes.Study,
    announcement: AnnouncementMsProfile,
):
    if "protocol" in type_map:
        for ref in type_map["protocol"]:
            protocol_parameters = []
            protocol_node: graph_nodes.Protocol = type_map["protocol"].get(ref)
            if not protocol_node.parameter_definition_refs:
                continue
            for definition_key in protocol_node.parameter_definition_refs:
                if definition_key not in all_nodes_map:
                    continue
                definition = all_nodes_map[definition_key]
                if not isinstance(definition, graph_nodes.ParameterDefinition):
                    continue
                vals = []
                for rel in relationship_name_map["has-instance"].values():
                    if rel.source_ref == definition.id_:
                        val = None
                        val_obj = all_nodes_map.get(rel.target_ref)
                        if isinstance(val_obj, CvTerm):
                            val = CvTerm.model_validate(
                                val_obj.model_dump(by_alias=True)
                            )
                        elif isinstance(val_obj, CvTermValue):
                            val = CvTermValue.model_validate(
                                val_obj.model_dump(by_alias=True)
                            )
                        if not val:
                            continue
                        vals.append(val)
                if vals:
                    def_type = all_nodes_map.get(definition.parameter_type_ref)
                    key = CvTerm.model_validate(def_type.model_dump(by_alias=True))
                    param = ExtendedCvTermKeyValue(
                        key=key,
                        values=vals if vals else None,
                    )
                    protocol_parameters.append(param)
            if not protocol_parameters:
                protocol_parameters = None
            else:
                protocol_parameters.sort(key=lambda x: x.key.name)
            protocol_type_object = all_nodes_map.get(
                protocol_node.protocol_type_ref, None
            )
            protocol = AnnouncementProtocol(
                name=protocol_node.name,
                protocol_type=protocol_type_object,
                description=protocol_node.description,
                protocol_parameters=protocol_parameters,
            )
            if not announcement.protocols:
                announcement.protocols = []
            announcement.protocols.append(protocol)


def convert_file(
    all_nodes_map,
    type_map: dict[str, dict[str, IdentifiableMhdModel]],
    file_type_name: str,
    ref: str,
    file_class: type[AnnouncementBaseFile],
):
    if file_type_name not in type_map or ref not in type_map[file_type_name]:
        return None
    item: graph_nodes.BaseFile = type_map[file_type_name][ref]
    url_list = item.url_list
    format = None
    if item.format_ref in all_nodes_map:
        format_node: BaseMhdModel = all_nodes_map[item.format_ref]
        format = CvTerm.model_validate(format_node.model_dump(by_alias=True))
    compressions = None
    if item.compression_format_refs in all_nodes_map:
        for format_ref in item.compression_format_refs:
            compression_node: BaseMhdModel = all_nodes_map[format_ref]
            compressions.append(
                CvTerm.model_validate(compression_node.model_dump(by_alias=True))
            )
    file = file_class(
        name=item.name,
        url_list=url_list,
        compression_formats=compressions,
        format=format,
    )

    return file


def collect_cvterm_sources(obj: BaseModel, cv_sources: set[str]):
    if isinstance(obj, (CvTerm, CvTermValue)):
        source = getattr(obj, "source", None)
        if source:
            cv_sources.add(source)
    elif isinstance(obj, BaseModel):
        for value in obj.__dict__.values():
            collect_cvterm_sources(value, cv_sources)
    elif isinstance(obj, list):
        for item in obj:
            collect_cvterm_sources(item, cv_sources)
    elif isinstance(obj, dict):
        for value in obj.values():
            collect_cvterm_sources(value, cv_sources)


def create_announcement_file(
    mhd_file: dict[str, Any],
    mhd_file_url: str,
    announcement_file_path: str,
    announcement_schema_name: str = ANNOUNCEMENT_FILE_V0_1_DEFAULT_SCHEMA_NAME,
    announcement_profile_uri=ANNOUNCEMENT_FILE_V0_1_LEGACY_PROFILE_NAME,
):
    mhd_dataset = MhDatasetMsProfile.model_validate(mhd_file)
    nodes_map = {x.id_: x for x in mhd_dataset.graph.nodes}
    relationships_map = {x.id_: x for x in mhd_dataset.graph.relationships}

    all_nodes_map: dict[str, BaseMhdModel] = {}
    type_map: dict[str, dict[str, BaseMhdModel]] = {}
    for node in mhd_dataset.graph.nodes:
        if node.type_ not in type_map:
            type_map[node.type_] = {}
        type_map[node.type_][node.id_] = node
        all_nodes_map[node.id_] = node

    relationship_name_map: dict[str, dict[str, BaseMhdRelationship]] = {}
    for rel in mhd_dataset.graph.relationships:
        if rel.relationship_name not in relationship_name_map:
            relationship_name_map[rel.relationship_name] = {}
        relationship_name_map[rel.relationship_name][rel.id_] = rel

    if "study" not in type_map:
        logger.info("Study not found for in input file")
        return
    study: graph_nodes.Study = list(type_map["study"].values())[0]

    study_assays: list[graph_nodes.Assay] = []
    if "assay" in type_map:
        study_assays = list(type_map["assay"].values())

    publications: list[AnnouncementPublication] = []
    if "publication" in type_map:
        graph_publications: list[graph_nodes.Publication] = list(
            type_map["publication"].values()
        )
        for node in graph_publications:
            item = AnnouncementPublication.model_validate(
                node.model_dump(by_alias=True)
            )
            publications.append(item)

    publication_status = None
    if not publications:
        if "defined-as" in relationship_name_map:
            publication_status = list(relationship_name_map["defined-as"].values())
            if publication_status:
                status = publication_status[0]

                publication_status = CvTerm.model_validate(
                    nodes_map[status.target_ref].model_dump(by_alias=True)
                )

    submitter_links: list[BaseMhdRelationship] = []
    if "submits" in relationship_name_map:
        submitter_links = list(relationship_name_map["submits"].values())

    submitters = []

    principal_investigators = []
    if "person" in type_map:
        for item in submitter_links:
            if item.source_ref in type_map["person"]:
                submitter = type_map["person"][item.source_ref]
                submitters.append(
                    AnnouncementContact.model_validate(submitter, from_attributes=True)
                )

        if "principal-investigator-of" in relationship_name_map:
            pi_links: list[BaseMhdRelationship] = list(
                relationship_name_map["principal-investigator-of"].values()
            )
            for item in pi_links:
                if item.source_ref in type_map["person"]:
                    pi = type_map["person"][item.source_ref]
                    principal_investigators.append(
                        AnnouncementContact.model_validate(pi, from_attributes=True)
                    )

    assay_types: OrderedDict[str, CvTerm] = OrderedDict()
    technology_types: OrderedDict[str, CvTerm] = OrderedDict()
    measurement_types: OrderedDict[str, CvTerm] = OrderedDict()
    for item in study_assays:
        if item.assay_type_ref in nodes_map:
            assay_type: graph_nodes.CvTermObject = nodes_map[item.assay_type_ref]
            if assay_type.accession not in assay_types:
                term = CvTerm.model_validate(assay_type)
                assay_types[term.accession] = term

        if item.technology_type_ref in nodes_map:
            technology_type: graph_nodes.CvTermObject = nodes_map[
                item.technology_type_ref
            ]
            if technology_type.accession not in technology_types:
                term = CvTerm.model_validate(technology_type)
                technology_types[term.accession] = term
        if item.measurement_type_ref in nodes_map:
            measurement_type: graph_nodes.CvTermObject = nodes_map[
                item.technology_type_ref
            ]
            if measurement_type.accession not in measurement_types:
                term = CvTerm.model_validate(technology_type)
                measurement_types[term.accession] = term

    dataset_url_list = study.url_list

    announcement = AnnouncementBaseProfile(
        mhd_identifier=study.mhd_identifier,
        repository_identifier=study.repository_identifier,
        schema_name=announcement_schema_name,
        profile_uri=announcement_profile_uri,
        mhd_metadata_file_url=CvTermValue(
            accession="EDAM:1052",
            name="URL",
            source="EDAM",
            value=mhd_file_url,
        ),
        dataset_url_list=dataset_url_list,
        license=study.license,
        title=study.title,
        description=study.description,
        submission_date=study.submission_date,
        public_release_date=study.public_release_date,
        submitters=submitters,
        principal_investigators=principal_investigators,
        measurement_type=list(measurement_types.values()),
        technology_type=list(technology_types.values()),
        assay_type=list(assay_types.values()),
        repository_metadata_file_list=[],
        result_file_list=[],
        raw_data_file_list=[],
        derived_data_file_list=[],
        supplementary_file_list=[],
        publications=publications if publications else publication_status,
        # study_factors=[],
        # characteristic_values=[],
    )

    update_keywords(all_nodes_map, relationship_name_map, announcement)
    update_protocol_parameters(
        all_nodes_map, relationship_name_map, type_map, study, announcement
    )
    update_study_factors(all_nodes_map, relationships_map, announcement)
    update_characteristic_values(all_nodes_map, relationships_map, announcement)

    if "metadata-file" in type_map:
        for ref in type_map["metadata-file"]:
            metadata = convert_file(
                all_nodes_map, type_map, "metadata-file", ref, AnnouncementMetadataFile
            )
            if metadata:
                announcement.repository_metadata_file_list.append(metadata)

    if "result-file" in type_map:
        for ref in type_map["result-file"]:
            file = convert_file(
                all_nodes_map, type_map, "result-file", ref, AnnouncementResultFile
            )
            if file:
                announcement.result_file_list.append(file)

    if "raw-data-file" in type_map:
        for ref in type_map["raw-data-file"]:
            file = convert_file(
                all_nodes_map, type_map, "raw-data-file", ref, AnnouncementRawDataFile
            )
            if file:
                announcement.raw_data_file_list.append(file)
    if "derived-data-file" in type_map:
        for ref in type_map["derived-data-file"]:
            file = convert_file(
                all_nodes_map,
                type_map,
                "derived-data-file",
                ref,
                AnnouncementDerivedDataFile,
            )
            if file:
                announcement.derived_data_file_list.append(file)
    if "supplementary-file" in type_map:
        for ref in type_map["supplementary-file"]:
            file = convert_file(
                all_nodes_map,
                type_map,
                "supplementary-file",
                ref,
                AnnouncementSupplementaryFile,
            )
            if file:
                announcement.supplementary_file_list.append(file)
    identification_map = {}
    identification_links = relationship_name_map.get("identified-as")
    items = type_map.get("metabolite-identification")
    if identification_links and items:
        for ref in identification_links:
            item = identification_links[ref]
            if item.target_ref in items:
                identification = items[item.target_ref]
                if item.source_ref not in identification_map:
                    identification_map[item.source_ref] = []
                identification_map[item.source_ref].append(identification)

    if "metabolite" in type_map:
        reported_metabolites = []
        for ref in type_map["metabolite"]:
            met = type_map["metabolite"][ref]
            item = AnnouncementReportedMetabolite(name=met.name)
            reported_metabolites.append(item)

            if ref in identification_map:
                identifications = identification_map[ref]
                item.database_identifiers = [
                    CvTermValue.model_validate(x.model_dump(by_alias=True))
                    for x in identifications
                ]

        if reported_metabolites:
            announcement.reported_metabolites = reported_metabolites
            announcement.reported_metabolites.sort(key=lambda x: x.name)
    cv_sources = set()
    collect_cvterm_sources(announcement, cv_sources)
    cv_sources = list(cv_sources)
    cv_sources.sort()
    for source in cv_sources:
        if source in CONTROLLED_CV_DEFINITIONS:
            announcement.cv_definitions.append(CONTROLLED_CV_DEFINITIONS[source])
        elif source in OTHER_CONTROLLED_CV_DEFINITIONS:
            announcement.cv_definitions.append(OTHER_CONTROLLED_CV_DEFINITIONS[source])
        else:
            announcement.cv_definitions.append(
                CvDefinition(label=source, alternative_labels=[source.lower()])
            )
    logger.info("Writing to %s", announcement_file_path)
    Path(announcement_file_path).parent.mkdir(parents=True, exist_ok=True)
    with Path(announcement_file_path).open("w") as f:
        f.write(
            announcement.model_dump_json(indent=2, by_alias=True, exclude_none=True)
        )



import pathlib
import sys

__version__ = "v0.1.31"

application_root_path = pathlib.Path(__file__).parent.parent

sys.path.append(str(application_root_path))

__all__ = [
    "domain_utils",
    "schema_utils",
    "utils",
    "schemas",
    "shared",
    "model",
    "convertors",
]



import logging
from typing import Any, Self, Sequence

from pydantic import Field
from typing_extensions import Annotated

from mhd_model.model.v0_1.dataset.profiles.base.base import (
    BaseLabeledMhdModel,
    BaseMhdRelationship,
    IdentifiableMhdModel,
    MhdObjectType,
)
from mhd_model.model.v0_1.dataset.profiles.base.profile import (
    GraphEnabledBaseDataset,
    MhDatasetBaseProfile,
)
from mhd_model.model.v0_1.dataset.profiles.base.relationships import Relationship
from mhd_model.model.v0_1.dataset.validation.utils import search_ontology_definition
from mhd_model.model.v0_1.rules.cv_definitions import (
    CONTROLLED_CV_DEFINITIONS,
    OTHER_CONTROLLED_CV_DEFINITIONS,
)
from mhd_model.shared.model import CvDefinition, CvTerm, CvTermValue

logger = logging.getLogger(__name__)


class MhDatasetBuilder(GraphEnabledBaseDataset):
    _cv_definitions_map: Annotated[
        dict[str, None | CvDefinition], Field(exclude=True)
    ] = {}

    type_: Annotated[MhdObjectType, Field(frozen=True, alias="type")] = MhdObjectType(
        "dataset"
    )

    objects: dict[str, IdentifiableMhdModel] = {}

    def add(self, item: IdentifiableMhdModel) -> Self:
        return self.add_node(item)

    def link(
        self,
        source: IdentifiableMhdModel,
        relationship_name: str,
        target: IdentifiableMhdModel,
        add_reverse_relationship: bool = False,
        reverse_relationship_name: None | str = None,
    ) -> Self:
        link = Relationship(
            source_ref=source.id_,
            relationship_name=relationship_name,
            target_ref=target.id_,
        )
        self.objects[link.id_] = link
        if add_reverse_relationship or reverse_relationship_name:
            reverse_relationship_name = (
                reverse_relationship_name
                if reverse_relationship_name
                else relationship_name
            )
            link = Relationship(
                source_ref=target.id_,
                relationship_name=reverse_relationship_name,
                target_ref=source.id_,
            )
            self.objects[link.id_] = link
        return self

    def add_node(self, item: IdentifiableMhdModel) -> Self:
        if item and isinstance(item, IdentifiableMhdModel) and item.id_:
            self.objects[item.id_] = item

            self.add_cv_source(item)
        else:
            logger.warning("Item %s is not valid. It will not be added.", item)
        return self

    def add_cv_source(self, item: Any) -> Self:
        if isinstance(item, (CvTerm, CvTermValue)):
            source_uppercase = item.source.upper() if item.source else ""
            if not source_uppercase:
                return self
            if source_uppercase not in self._cv_definitions_map:
                logger.info("%s CV source is added.", item.source)
                self._cv_definitions_map[source_uppercase] = None
            if source_uppercase != item.source:
                logger.info(
                    "CV source '%s' is converted to uppercase '%s'. %s",
                    item.source,
                    source_uppercase,
                    item,
                )
                item.source = source_uppercase
        return self

    def add_relationship(self, item: BaseMhdRelationship) -> Self:
        self.objects[item.id_] = item
        return self

    def create_dataset(
        self, start_item_refs: Sequence[str], dataset_class: type[MhDatasetBaseProfile]
    ) -> MhDatasetBaseProfile:
        cv_definitions_map: dict[str, CvDefinition] = {}

        for source in self._cv_definitions_map.keys():
            if not source:
                continue
            if source in CONTROLLED_CV_DEFINITIONS:
                cv_definition = CONTROLLED_CV_DEFINITIONS[source]
                self.cv_definitions.append(cv_definition)
                cv_definitions_map[source] = cv_definition
            elif source in OTHER_CONTROLLED_CV_DEFINITIONS:
                cv_definition = OTHER_CONTROLLED_CV_DEFINITIONS[source]
                self.cv_definitions.append(cv_definition)
                cv_definitions_map[source] = cv_definition
            else:
                cv_definition = search_ontology_definition(source)
                if not cv_definition:
                    self.cv_definitions.append(CvDefinition(label=source))
                else:
                    self.cv_definitions.append(cv_definition)
                cv_definitions_map[source] = cv_definition

        self.cv_definitions.sort(key=lambda x: x.label)
        mhd_dataset = dataset_class(
            schema_name=self.schema_name, profile_uri=self.profile_uri
        )
        mhd_dataset.cv_definitions = (
            self.cv_definitions.copy() if self.cv_definitions else []
        )
        mhd_dataset.repository_name = self.repository_name
        mhd_dataset.revision = self.revision
        mhd_dataset.repository_identifier = self.repository_identifier
        mhd_dataset.mhd_identifier = self.mhd_identifier
        mhd_dataset.revision_datetime = self.revision_datetime
        mhd_dataset.repository_revision = self.repository_revision
        mhd_dataset.repository_revision_datetime = self.repository_revision_datetime
        mhd_dataset.change_log = self.change_log.copy() if self.change_log else None

        iterated_items: set[str] = set()
        for identifier, item in self.objects.items():
            if identifier not in iterated_items:
                iterated_items.add(identifier)
                if identifier in start_item_refs:
                    mhd_dataset.graph.start_item_refs.append(identifier)
                if isinstance(item, BaseMhdRelationship):
                    mhd_dataset.graph.relationships.append(item)
                else:
                    mhd_dataset.graph.nodes.append(item)

        def sort_key(item: BaseLabeledMhdModel):
            if isinstance(item, CvTerm):
                return (100, item.type_, item.label, item.id_)
            if item.id_ in start_item_refs:
                return (0, item.type_, item.label, item.id_)
            if isinstance(item, BaseMhdRelationship):
                return (
                    0,
                    item.source_ref,
                    item.relationship_name,
                    item.target_ref,
                    item.id_,
                )
            if item.id_.startswith("cv-"):
                return (100, item.type_, item.label, item.id_)
            return (2, item.type_, item.label, item.id_)

        mhd_dataset.graph.nodes = sorted(mhd_dataset.graph.nodes, key=sort_key)
        mhd_dataset.graph.relationships.sort(key=sort_key)
        return mhd_dataset

    @classmethod
    def from_dataset(cls, mhd_dataset: MhDatasetBaseProfile) -> "MhDatasetBuilder":
        dataset = cls(
            schema_name=mhd_dataset.schema_name, profile_uri=mhd_dataset.profile_uri
        )
        dataset.cv_definitions = (
            mhd_dataset.cv_definitions.copy() if mhd_dataset.cv_definitions else []
        )
        dataset.repository_name = mhd_dataset.repository_name
        dataset.mhd_identifier = mhd_dataset.mhd_identifier
        dataset.repository_identifier = mhd_dataset.repository_identifier
        dataset.revision = mhd_dataset.revision
        dataset.revision_datetime = mhd_dataset.revision_datetime
        dataset.repository_revision = mhd_dataset.repository_revision
        dataset.repository_revision_datetime = mhd_dataset.repository_revision_datetime
        dataset.change_log = (
            mhd_dataset.change_log.copy() if mhd_dataset.change_log else []
        )

        for item in mhd_dataset.graph.nodes:
            dataset.objects[item.id_] = item
        for item in mhd_dataset.graph.relationships:
            dataset.objects[item.id_] = item
        return dataset



import json
import pathlib
from typing import Any


def load_json(file: str) -> dict[str, Any]:
    with pathlib.Path(file).open("r") as f:
        json_file = json.load(f)
    return json_file


def json_path(field_path: list[int | str]) -> str:
    return ".".join([x if isinstance(x, str) else f"[{x}]" for x in field_path])



import json
import logging
from pathlib import Path
from typing import OrderedDict

from mhd_model.model.v0_1.announcement.profiles.base.fields import (
    ExtendedCvTermKeyValue,
)
from mhd_model.model.v0_1.announcement.profiles.base.profile import (
    AnnouncementBaseFile,
    AnnouncementProtocol,
)
from mhd_model.model.v0_1.dataset.profiles.base import graph_nodes
from mhd_model.model.v0_1.dataset.profiles.base.base import (
    BaseMhdModel,
    BaseMhdRelationship,
    CvTerm,
    CvTermValue,
    IdentifiableMhdModel,
)
from mhd_model.model.v0_1.dataset.profiles.legacy.profile import MhDatasetLegacyProfile
from mhd_model.model.v0_1.rules.managed_cv_terms import (
    COMMON_ASSAY_TYPES,
    COMMON_MEASUREMENT_TYPES,
    COMMON_OMICS_TYPES,
    COMMON_TECHNOLOGY_TYPES,
)
from mhd_model.model.v0_1.sdrf.model import (
    LegacySdrf,
    LegacySdrfRow,
    SdrfFile,
    SdrfKeyValue,
    SdrfSampleProtocolDefinition,
)
from mhd_model.shared.model import QuantitativeValue

logger = logging.getLogger(__name__)


def get_characteristic_values(
    all_nodes_map: dict[str, IdentifiableMhdModel],
    relationships_map: dict[str, BaseMhdRelationship],
) -> list[ExtendedCvTermKeyValue]:
    study_characteristics = set()
    characteristic_values: OrderedDict[str, list[str]] = OrderedDict()
    for rel in relationships_map.values():
        source = all_nodes_map[rel.source_ref]
        if rel.relationship_name == "has-characteristic-definition":
            if isinstance(source, graph_nodes.Study):
                study_characteristics.add(rel.target_ref)

        if rel.relationship_name == "has-instance":
            if isinstance(source, graph_nodes.CharacteristicDefinition):
                if rel.source_ref not in characteristic_values:
                    characteristic_values[rel.source_ref] = []
                characteristic_values[rel.source_ref].append(rel.target_ref)
    referenced_characteristics = {
        x: y
        for x, y in characteristic_values.items()
        if x in study_characteristics and y
    }
    characteristic_keys = [
        (x, all_nodes_map[x]) for x, y in referenced_characteristics.items()
    ]
    characteristic_keys.sort(key=lambda x: x[1].name)
    announcement_characteristics = []
    for char_id, characteristic in characteristic_keys:
        value_ids = referenced_characteristics[char_id]
        values = []
        for x in value_ids:
            val = all_nodes_map[x]
            if val.id_.startswith("cv-value--"):
                val = CvTermValue.model_validate(val.model_dump(by_alias=True))
            else:
                val = CvTerm.model_validate(val.model_dump(by_alias=True))
            values.append(val)
        type_node = all_nodes_map.get(characteristic.characteristic_type_ref, None)
        key = CvTerm.model_validate(type_node.model_dump(by_alias=True))
        announcement_characteristics.append(
            ExtendedCvTermKeyValue(key=key, values=values)
        )

    return announcement_characteristics


def get_study_factors(
    all_nodes_map: dict[str, IdentifiableMhdModel],
    relationships_map: dict[str, BaseMhdRelationship],
):
    study_factors = set()
    factors: OrderedDict[str, list[str]] = OrderedDict()
    for rel in relationships_map.values():
        source = all_nodes_map[rel.source_ref]
        if rel.relationship_name == "has-factor-definition":
            if isinstance(source, graph_nodes.Study):
                study_factors.add(rel.target_ref)

        if rel.relationship_name == "has-instance":
            if isinstance(source, graph_nodes.FactorDefinition):
                if rel.source_ref not in factors:
                    factors[rel.source_ref] = []
                factors[rel.source_ref].append(rel.target_ref)
    referenced_factors = {x: y for x, y in factors.items() if x in study_factors and y}
    factor_keys = [(x, all_nodes_map[x]) for x, y in referenced_factors.items()]
    factor_keys.sort(key=lambda x: x[1].name)
    announcement_factors = []
    for factor_id, factor in factor_keys:
        value_ids = referenced_factors[factor_id]
        values = []
        for x in value_ids:
            val = all_nodes_map[x]
            if isinstance(val, CvTermValue):
                val = CvTermValue.model_validate(val.model_dump(by_alias=True))
            else:
                val = CvTerm.model_validate(val.model_dump(by_alias=True))
            values.append(val)
        type_node = all_nodes_map.get(factor.factor_type_ref, None)
        key = CvTerm.model_validate(type_node.model_dump(by_alias=True))
        announcement_factors.append(ExtendedCvTermKeyValue(key=key, values=values))
    return announcement_factors


def get_protocols(
    all_nodes_map: dict[str, IdentifiableMhdModel],
    relationship_name_map: dict[str, BaseMhdRelationship],
    type_map: dict[str, dict[IdentifiableMhdModel]],
    study: graph_nodes.Study,
):
    protocols: list[AnnouncementProtocol] = []
    if not study.protocol_refs:
        return protocols

    for ref in study.protocol_refs:
        if ref not in type_map["protocol"]:
            logger.error("Protocol %s is not defined in dataset.", ref)
        protocol_parameters = []
        protocol_node: graph_nodes.Protocol = type_map["protocol"].get(ref)
        protocol_type_object = all_nodes_map.get(protocol_node.protocol_type_ref, None)
        # if not protocol_node.parameter_definition_refs:
        #     continue
        protocol_parameters = []
        if protocol_node.parameter_definition_refs:
            for definition_key in protocol_node.parameter_definition_refs:
                if definition_key not in all_nodes_map:
                    continue
                definition = all_nodes_map[definition_key]
                if not isinstance(definition, graph_nodes.ParameterDefinition):
                    continue
                vals = []
                for rel in relationship_name_map["has-instance"].values():
                    if rel.source_ref == definition.id_:
                        val = None
                        val_obj = all_nodes_map.get(rel.target_ref)
                        if isinstance(val_obj, CvTerm):
                            val = CvTerm.model_validate(
                                val_obj.model_dump(by_alias=True)
                            )
                        elif isinstance(val_obj, CvTermValue):
                            val = CvTermValue.model_validate(
                                val_obj.model_dump(by_alias=True)
                            )
                        if not val:
                            continue
                        vals.append(val)
                # if vals:
                def_type = all_nodes_map.get(definition.parameter_type_ref)
                key = CvTerm.model_validate(def_type.model_dump(by_alias=True))
                param = ExtendedCvTermKeyValue(
                    key=key,
                    values=vals or None,
                )
                protocol_parameters.append(param)
        if not protocol_parameters:
            protocol_parameters = None
        else:
            protocol_parameters.sort(key=lambda x: x.key.name)
        protocol_type = CvTerm.model_validate(
            protocol_type_object.model_dump(by_alias=True)
        )
        protocol = AnnouncementProtocol(
            name=protocol_node.name,
            protocol_type=protocol_type,
            description=protocol_node.description,
            protocol_parameters=protocol_parameters,
        )
        protocols.append(protocol)

    return protocols


def convert_file(
    all_nodes_map,
    type_map: dict[str, dict[str, IdentifiableMhdModel]],
    file_type_name: str,
    ref: str,
    file_class: type[AnnouncementBaseFile],
):
    if file_type_name not in type_map or ref not in type_map[file_type_name]:
        return None
    item: graph_nodes.BaseFile = type_map[file_type_name][ref]
    url_list = item.url_list
    format = None
    if item.format_ref in all_nodes_map:
        format_node: BaseMhdModel = all_nodes_map[item.format_ref]
        format = CvTerm.model_validate(format_node.model_dump(by_alias=True))
    compressions = None
    if item.compression_format_refs in all_nodes_map:
        for format_ref in item.compression_format_refs:
            compression_node: BaseMhdModel = all_nodes_map[format_ref]
            compressions.append(
                CvTerm.model_validate(compression_node.model_dump(by_alias=True))
            )
    file = file_class(
        name=item.name,
        url_list=url_list,
        compression_formats=compressions,
        format=format,
    )

    return file


def create_sdrf_files(
    mhd_file_path: str,
    sdrf_file_root_path: str,
    assay_name: None | str = None,
    sdrf_output_filename: None | str = None,
):
    sdrf_files = []
    txt = Path(mhd_file_path).read_text()
    mhd_data_json = json.loads(txt)

    mhd_dataset = MhDatasetLegacyProfile.model_validate(mhd_data_json)
    nodes_map: dict[str, IdentifiableMhdModel] = {
        x.id_: x for x in mhd_dataset.graph.nodes
    }
    # relationships_map: dict[str, BaseMhdRelationship] = {
    #     x.id_: x for x in mhd_dataset.graph.relationships
    # }

    all_nodes_map: dict[str, BaseMhdModel] = {}
    type_map: dict[str, dict[str, BaseMhdModel]] = {}
    for node in mhd_dataset.graph.nodes:
        if node.type_ not in type_map:
            type_map[node.type_] = {}
        type_map[node.type_][node.id_] = node
        all_nodes_map[node.id_] = node

    relationship_name_map: dict[str, dict[str, BaseMhdRelationship]] = {}
    node_relationships: dict[str, dict[str, list[str]]] = {}
    for rel in mhd_dataset.graph.relationships:
        if rel.relationship_name not in relationship_name_map:
            relationship_name_map[rel.relationship_name] = {}
        relationship_name_map[rel.relationship_name][rel.id_] = rel
        if rel.source_ref not in node_relationships:
            node_relationships[rel.source_ref] = {}
        if rel.relationship_name not in node_relationships[rel.source_ref]:
            node_relationships[rel.source_ref][rel.relationship_name] = []
        node_relationships[rel.source_ref][rel.relationship_name].append(rel.target_ref)

    if "study" not in type_map or not type_map["study"]:
        logger.error("Study is not found for in the input file")
        return False, sdrf_files
    study: graph_nodes.Study = list(type_map["study"].values())[0]
    study_id = study.mhd_identifier or study.repository_identifier
    study_assays: list[graph_nodes.Assay] = []
    if "assay" in type_map:
        study_assays = list(type_map["assay"].values())
    if not study_assays:
        logger.error("Study %s has no assay", study_id)
        return False, sdrf_files
    selected_assays = [(idx, x) for idx, x in enumerate(study_assays, start=1)]
    if assay_name:
        selected_assays = [
            (idx, x)
            for idx, x in enumerate(study_assays, start=1)
            if x.name == assay_name
        ]
    if not selected_assays:
        logger.error("Study %s has no assay with name %s", study_id, assay_name)
        return False, sdrf_files
    for idx, assay in selected_assays:
        if not assay.sample_run_refs:
            logger.error("Study %s Assay '%s' has no sample run", study_id, assay.name)
            continue
        sdrf_file = LegacySdrf(
            study_identifier=study.repository_identifier,
            assay_identifier=assay.repository_identifier,
        )
        parameter_type_cv_terms: dict[str, CvTerm] = {}

        for sample_run_ref in assay.sample_run_refs:
            sample_run: graph_nodes.SampleRun = all_nodes_map[sample_run_ref]
            sample: graph_nodes.Sample = all_nodes_map[sample_run.sample_ref]

            subjects = find_all_linked_nodes(
                all_nodes_map,
                node_relationships,
                sample.id_,
                "derived-from",
                "subject",
            )
            subject = subjects[0] if subjects else None
            if subject:
                values: list[IdentifiableMhdModel] = find_all_linked_nodes(
                    all_nodes_map,
                    node_relationships,
                    subject.id_,
                    "has-characteristic-value",
                    "characteristic-value",
                )
                characteristic_values = []
                if values:
                    for value in values:
                        characteristic_types = find_all_linked_nodes(
                            all_nodes_map,
                            node_relationships,
                            value.id_,
                            "has-type",
                            "characteristic-type",
                        )
                        characteristic_type = (
                            characteristic_types[0] if characteristic_types else None
                        )
                        if characteristic_type:
                            target_value = None
                            if value.value is not None and value.name is None:
                                target_value = QuantitativeValue.model_validate(
                                    value.model_dump(by_alias=True)
                                )
                            else:
                                target_value = CvTerm.model_validate(
                                    value.model_dump(by_alias=True)
                                )
                            characteristic_values.append(
                                SdrfKeyValue(
                                    key=CvTerm.model_validate(
                                        characteristic_type.model_dump(by_alias=True)
                                    ),
                                    value=target_value,
                                )
                            )
                values: list[IdentifiableMhdModel] = find_all_linked_nodes(
                    all_nodes_map,
                    node_relationships,
                    sample.id_,
                    "has-factor-value",
                    "factor-value",
                )
                factor_values = []
                if values:
                    for value in values:
                        # select only first type
                        factor_types = find_all_linked_nodes(
                            all_nodes_map,
                            node_relationships,
                            value.id_,
                            "has-type",
                            "factor-type",
                        )
                        factor_type = factor_types[0] if factor_types else None
                        if factor_type:
                            target_value = None
                            if value.value is not None and value.name is None:
                                target_value = QuantitativeValue.model_validate(
                                    value.model_dump(by_alias=True)
                                )
                            else:
                                target_value = CvTerm.model_validate(
                                    value.model_dump(by_alias=True)
                                )
                            factor_values.append(
                                SdrfKeyValue(
                                    key=CvTerm.model_validate(
                                        factor_type.model_dump(by_alias=True)
                                    ),
                                    value=target_value,
                                )
                            )
            raw_data_files = []
            derived_data_files = []
            result_files = []
            supplementary_files = []
            if sample_run.raw_data_file_refs:
                raw_data_file_nodes: graph_nodes.RawDataFile = [
                    nodes_map[x] for x in sample_run.raw_data_file_refs
                ]
                raw_data_files = [
                    SdrfFile(name=x.name, url_list=x.url_list if x.url_list else None)
                    for x in raw_data_file_nodes
                ]
            if sample_run.derived_data_file_refs:
                derived_data_file_nodes: graph_nodes.DerivedDataFile = [
                    nodes_map[x] for x in sample_run.derived_data_file_refs
                ]
                derived_data_files = [
                    SdrfFile(name=x.name, url_list=x.url_list if x.url_list else None)
                    for x in derived_data_file_nodes
                ]
            if sample_run.result_file_refs:
                result_file_nodes: graph_nodes.ResultFile = [
                    nodes_map[x] for x in sample_run.result_file_refs
                ]
                result_files = [
                    SdrfFile(name=x.name, url_list=x.url_list if x.url_list else None)
                    for x in result_file_nodes
                ]
            if sample_run.supplementary_file_refs:
                supplementary_file_nodes: graph_nodes.SupplementaryFile = [
                    nodes_map[x] for x in sample_run.supplementary_file_refs
                ]
                supplementary_files = [
                    SdrfFile(name=x.name, url_list=x.url_list if x.url_list else None)
                    for x in supplementary_file_nodes
                ]
            protocol_parameters = []
            if sample_run.sample_run_configuration_refs:
                run_configs: graph_nodes.SampleRunConfiguration = [
                    nodes_map[x] for x in sample_run.sample_run_configuration_refs
                ]
                for run_config in run_configs:
                    protocol = nodes_map[run_config.protocol_ref]
                    protocol_type = nodes_map[protocol.protocol_type_ref]
                    protocol_name = protocol.name
                    if run_config.parameter_value_refs:
                        parameter_value_nodes = [
                            nodes_map[x] for x in run_config.parameter_value_refs
                        ]
                        parameter_values = []
                        for value in parameter_value_nodes:
                            paramter_types: list[IdentifiableMhdModel] = (
                                find_all_linked_nodes(
                                    all_nodes_map,
                                    node_relationships,
                                    value.id_,
                                    "has-type",
                                    "parameter-type",
                                )
                            )
                            if paramter_types:
                                parameter_type = paramter_types[0]
                                if parameter_type.id_ not in parameter_type_cv_terms:
                                    parameter_type_cv_terms[parameter_type.id_] = (
                                        CvTerm.model_validate(
                                            parameter_type.model_dump(by_alias=True)
                                        )
                                    )
                                target_value = None
                                if value.value is not None and value.name is None:
                                    target_value = QuantitativeValue.model_validate(
                                        value.model_dump(by_alias=True)
                                    )
                                else:
                                    target_value = CvTerm.model_validate(
                                        value.model_dump(by_alias=True)
                                    )
                                parameter_values.append(
                                    SdrfKeyValue(
                                        key=parameter_type_cv_terms[parameter_type.id_],
                                        value=target_value,
                                    )
                                )

                    protocol_parameters.append(
                        SdrfSampleProtocolDefinition(
                            protocol_name=protocol_name,
                            protocol_type=CvTerm.model_validate(
                                protocol_type.model_dump(by_alias=True)
                            ),
                            parameter_values=parameter_values,
                        )
                    )
            row = LegacySdrfRow(
                assay_name=sample_run.name or "",
                sample_name=sample.name or "",
                characteristics=characteristic_values,
                factors=factor_values,
                raw_data_files=raw_data_files,
                derived_data_files=derived_data_files,
                result_files=result_files,
                supplementary_files=supplementary_files,
                protocol_parameters=protocol_parameters,
            )
            sdrf_file.sample_runs.append(row)

        headers: OrderedDict[str, str] = OrderedDict()
        characteristic_definitions: list[IdentifiableMhdModel] = find_all_linked_nodes(
            all_nodes_map,
            node_relationships,
            study.id_,
            "has-characteristic-definition",
            "characteristic-definition",
        )
        headers["sample name"] = None
        factor_headers = {}
        characteristic_headers = {}
        parameter_value_headers = {}

        characteristic_definitions.sort(key=lambda x: x.name)
        for definition in characteristic_definitions:
            definition_type = nodes_map[definition.characteristic_type_ref]
            name = (
                definition_type.name.lower()
                if definition_type.name
                else definition.name.lower()
            )
            if definition_type.accession:
                header = f"characteristic[{name}|{definition_type.accession}]"
                headers[header] = definition
            else:
                header = f"characteristic[{name}]"
                headers[header] = definition
            characteristic_headers[name] = header

        headers["assay name"] = None
        if assay.protocol_refs:
            protocols = [nodes_map[x] for x in assay.protocol_refs]
            for protocol_item in protocols:
                parameter_definitions: list[IdentifiableMhdModel] = (
                    find_all_linked_nodes(
                        all_nodes_map,
                        node_relationships,
                        protocol_item.id_,
                        "has-parameter-definition",
                        "parameter-definition",
                    )
                )
                if parameter_definitions:
                    for param in parameter_definitions:
                        parameter_type = nodes_map[param.parameter_type_ref]
                        name = (
                            parameter_type.name.lower()
                            if parameter_type.name
                            else param.name.lower()
                        )
                        if parameter_type and parameter_type.accession:
                            header = (
                                f"parameter value[{name}|{parameter_type.accession}]"
                            )
                            headers[header] = definition
                        else:
                            header = f"parameter value[{name}]"
                            headers[header] = definition
                        parameter_value_headers[name] = header

        max_raw_data_column = 0
        max_derived_data_column = 0
        max_result_data_column = 0
        max_supplementary_data_column = 0
        for row in sdrf_file.sample_runs:
            if row.raw_data_files:
                max_raw_data_column = max(max_raw_data_column, len(row.raw_data_files))
            if row.derived_data_files:
                max_derived_data_column = max(
                    max_derived_data_column, len(row.derived_data_files)
                )
            if row.result_files:
                max_result_data_column = max(
                    max_result_data_column, len(row.result_files)
                )
            if row.supplementary_files:
                max_supplementary_data_column = max(
                    max_supplementary_data_column, len(row.supplementary_files)
                )
        raw_files_data: list[list[tuple[str, str]]] = [[]] * max_raw_data_column
        derived_files_data: list[list[tuple[str, str]]] = [[]] * max_derived_data_column
        result_files_data: list[list[tuple[str, str]]] = [[]] * max_result_data_column
        summplementary_files_data: list[list[tuple[str, str]]] = [
            []
        ] * max_supplementary_data_column
        for name_prefix, files in [
            ("raw data file", raw_files_data),
            ("derived data file", derived_files_data),
            ("result file", result_files_data),
            ("supplementary file", summplementary_files_data),
        ]:
            for idx, item in enumerate(files, start=1):
                name = f"comment[{name_prefix} name.{idx}]"
                headers[name] = name
                url = f"comment[{name_prefix} url.{idx}]"
                headers[url] = url
        factor_definitions: list[IdentifiableMhdModel] = find_all_linked_nodes(
            all_nodes_map,
            node_relationships,
            study.id_,
            "has-factor-definition",
            "factor-definition",
        )
        factor_definitions.sort(key=lambda x: x.name)
        for definition in factor_definitions:
            definition_type = nodes_map[definition.factor_type_ref]
            name = (
                definition_type.name.lower()
                if definition_type.name
                else definition.name.lower()
            )
            if definition_type.accession:
                header = f"factor value[{name}|{definition_type.accession}]"
                headers[header] = definition
            else:
                header = f"factor value[{name}]"
                headers[header] = definition
            factor_headers[name] = header

        tsv_data: OrderedDict[str, list[str]] = OrderedDict([(x, []) for x in headers])

        for row in sdrf_file.sample_runs:
            tsv_data.get("sample name", []).append(row.sample_name or "")
            tsv_data.get("assay name", []).append(row.assay_name or "")
            filled_headers = set(["sample name", "assay name"])
            all_param_value_definitions = []
            for item in row.protocol_parameters:
                all_param_value_definitions.extend(item.parameter_values)

            for items, item_headers in [
                (row.characteristics, characteristic_headers),
                (row.factors, factor_headers),
                (all_param_value_definitions, parameter_value_headers),
            ]:
                for item in items:
                    if item.key and item.key.name:
                        item_type = item.key.name
                        header = item_headers.get(item_type)
                        filled_headers.add(header)
                        tsv_data.get(header, []).append(item.value.name or "")

            for name_prefix, files, values in [
                ("raw data file", raw_files_data, row.raw_data_files),
                ("derived data file", derived_files_data, row.derived_data_files),
                ("result file", result_files_data, row.result_files),
                (
                    "supplementary file",
                    summplementary_files_data,
                    row.supplementary_files,
                ),
            ]:
                for idx, item in enumerate(files, start=1):
                    header = f"comment[{name_prefix} name.{idx}]"
                    name_value = values[idx - 1].name or ""
                    tsv_data.get(header, []).append(name_value or "")
                    url_header = f"comment[{name_prefix} url.{idx}]"
                    url_list = values[idx - 1].url_list
                    filled_headers.add(header)
                    url_value = ""
                    if url_list:
                        url_value = str(url_list[0])
                    tsv_data.get(url_header, []).append(url_value)
                    filled_headers.add(url_header)
            not_defined_headers = {x for x in tsv_data if x not in filled_headers}
            for header in not_defined_headers:
                tsv_data[header].append("")
        empty_columns = []
        exceptions = {
            "sample name",
            "assay name",
            "comment[raw data file name.1]",
            "comment[raw data file url.1]",
        }
        for header, values in tsv_data.items():
            if header in exceptions:
                continue
            unique_values = set(values)
            unique_values.discard("")
            unique_values.discard(None)
            if not unique_values:
                empty_columns.append(header)
        for item in empty_columns:
            del tsv_data[item]
        if empty_columns:
            logger.debug("Empty columns are dropped: %s", ", ".join(empty_columns))
        # assay_id = sdrf_file.assay_identifier

        if sdrf_output_filename and assay_name:
            sdrf_filename = sdrf_output_filename
        else:
            if assay.name.startswith(f"{study_id}_"):
                sdrf_filename = f"{Path(assay.name).stem}.sdrf.tsv"
            else:
                sdrf_filename = f"{study_id}_{Path(assay.name).stem}.sdrf.tsv"
        srdf_file_path = Path(sdrf_file_root_path) / Path(sdrf_filename)
        srdf_file_path.parent.mkdir(parents=True, exist_ok=True)
        sdrf_files.append(srdf_file_path)
        with srdf_file_path.open("w") as f:
            headers = "\t".join(list(tsv_data.keys())) + "\n"
            f.write(headers)
            row_count = len(tsv_data["sample name"])
            for idx in range(row_count):
                row = [tsv_data[x][idx] for x in tsv_data]
                row_data = "\t".join(row) + "\n"
                f.write(row_data)
        logger.info("%s SDRF file is created: %s", study_id, str(srdf_file_path))

    return sdrf_files


def find_all_linked_nodes(
    all_nodes_map: dict[str, IdentifiableMhdModel],
    node_relationships: dict[str, dict[str, list[str]]],
    source_id: str,
    relationship_name: str,
    target_node_type: str,
):
    links = node_relationships.get(source_id, {}).get(relationship_name, [])
    nodes = [
        all_nodes_map[x]
        for x in links
        if all_nodes_map[x].type_ == target_node_type
        or (
            all_nodes_map[x].type_.startswith("x-")
            and all_nodes_map[x].type_.endswith(target_node_type)
        )
    ]
    return nodes


def get_file_list(
    all_nodes_map: dict[str, IdentifiableMhdModel],
    type_map: dict[str, dict[str, BaseMhdModel]],
    node_type: str,
    file_class: type[AnnouncementBaseFile],
) -> list[AnnouncementBaseFile]:
    file_list = []
    if node_type in type_map:
        for ref in type_map[node_type]:
            file_object = convert_file(
                all_nodes_map, type_map, node_type, ref, file_class
            )
            if file_object:
                file_list.append(file_object)
    return file_list


def get_main_assay_descriptors(
    nodes_map: dict[str, IdentifiableMhdModel],
    study_assays: list[graph_nodes.Assay],
    keywords: list[CvTerm],
) -> tuple[list[CvTerm], list[CvTerm], list[CvTerm], list[CvTerm]]:
    assay_types: dict[str, CvTerm] = {}
    technology_types: dict[str, CvTerm] = {}
    measurement_types: dict[str, CvTerm] = {}
    omics_types: dict[str, CvTerm] = {}
    for item in study_assays:
        if item.assay_type_ref in nodes_map:
            assay_type: graph_nodes.CvTermObject = nodes_map[item.assay_type_ref]
            if assay_type.accession not in assay_types:
                term = COMMON_ASSAY_TYPES.get(assay_type.accession, None)
                if not term:
                    term = CvTerm.model_validate(assay_type.model_dump(by_alias=True))
                assay_types[term.accession] = term

        if item.technology_type_ref in nodes_map:
            technology_type: graph_nodes.CvTermObject = nodes_map[
                item.technology_type_ref
            ]
            if technology_type.accession not in technology_types:
                term = COMMON_TECHNOLOGY_TYPES.get(technology_type.accession)
                if not term:
                    term = CvTerm.model_validate(
                        technology_type.model_dump(by_alias=True)
                    )
                technology_types[term.accession] = term
        if item.measurement_type_ref in nodes_map:
            measurement_type: graph_nodes.CvTermObject = nodes_map[
                item.technology_type_ref
            ]
            if measurement_type.accession not in measurement_types:
                term = COMMON_MEASUREMENT_TYPES.get(measurement_type.accession)
                if not term:
                    term = CvTerm.model_validate(
                        technology_type.model_dump(by_alias=True)
                    )
                measurement_types[term.accession] = term

    for keyword in keywords:
        if "untargetted" in keyword.name.lower():
            measurement_types["MSIO:0000101"] = COMMON_MEASUREMENT_TYPES["MSIO:0000101"]
        elif "targetted" in keyword.name.lower():
            measurement_types["MSIO:0000100"] = COMMON_MEASUREMENT_TYPES["MSIO:0000101"]
        for accession, value in COMMON_OMICS_TYPES.items():
            if accession.lower() == keyword.name.lower():
                omics_types[accession] = value

    return (
        list(assay_types.values()),
        list(technology_types.values()),
        list(measurement_types.values()),
        list(omics_types.values()),
    )



import json
import logging
import pathlib

from mhd_model.model.v0_1.announcement.profiles.base.profile import (
    Announcement,
)
from mhd_model.model.v0_1.announcement.profiles.ms.profile import (
    AnnouncementMsProfile,
)

logger = logging.getLogger(__name__)


def update_annoucement_profiles():
    profile_path = "mhd_model/schemas/mhd/announcement-v0.1.schema.json"
    with pathlib.Path(profile_path).open("w") as f:
        json.dump(Announcement.model_json_schema(), f, indent=2)
    logger.info(
        "Base announcement profile file on directory '%s' is updated.",
        profile_path,
    )

    profile_path = "mhd_model/schemas/mhd/announcement-v0.1.schema.ms-profile.json"
    with pathlib.Path(profile_path).open("w") as f:
        json.dump(AnnouncementMsProfile.model_json_schema(), f, indent=2)
    logger.info(
        "MS announcement profile file on directory '%s' is updated.",
        profile_path,
    )



from mhd_model.shared.model import CvTerm

PREDEFINED_CV_TERMS: dict[str, dict[str, CvTerm]] = {
    "OBI:0000470": CvTerm(
        source="OBI",
        accession="OBI:0000470",
        name="mass spectrometry assay",
    ),
    "OBI:0000623": CvTerm(
        source="OBI",
        accession="OBI:0000623",
        name="NMR spectroscopy assay",
    ),
    "OBI:0003097S": CvTerm(
        source="OBI",
        accession="OBI:0003097",
        name="liquid chromatography mass spectrometry assay",
    ),
    "OBI:0003110": CvTerm(
        source="OBI",
        accession="OBI:0003110",
        name="gas chromatography mass spectrometry assay",
    ),
    "OBI:0000366": CvTerm(
        source="OBI",
        accession="OBI:0000366",
        name="metabolite profiling assay",
    ),
    "OBI:0000485": CvTerm(
        source="OBI", accession="OBI:0000485", name="chromatography instrument"
    ),
    "OBI:0000521": CvTerm(
        source="OBI", accession="OBI:0000521", name="flame ionization detector"
    ),
    "OBI:0000345": CvTerm(source="OBI", accession="OBI:0000345", name="mass analyzer"),
    "OBI:0001132": CvTerm(
        source="OBI",
        accession="OBI:0001132",
        name="capillary electrophoresis instrument",
    ),
    "OBI:0200000": CvTerm(
        source="OBI",
        accession="OBI:0200000",
        name="data transform",
    ),
    "MS:1000031": CvTerm(
        source="MS",
        accession="MS:1000031",
        name="instrument model",
    ),
    "MS:1000560": CvTerm(
        source="MS",
        accession="MS:1000560",
        name="mass spectrometer file format",
    ),
    "MS:1002853": CvTerm(
        source="MS",
        accession="MS:1002853",
        name="no associated published manuscript",
    ),
    "MS:1002858": CvTerm(
        source="MS",
        accession="MS:1002858",
        name="Dataset with its publication pending",
    ),
    "NCIT:C14250": CvTerm(source="NCIT", accession="NCIT:C14250", name="organism"),
    "NCIT:C103199": CvTerm(
        source="NCIT", accession="NCIT:C103199", name="organism part"
    ),
    "EFO:0000324": CvTerm(source="EFO", accession="EFO:0000324", name="cell type"),
    "EFO:0000408": CvTerm(source="EFO", accession="EFO:0000408", name="disease"),
    "CHMO:0000470": CvTerm(
        source="CHMO",
        accession="CHMO:0000470",
        name="mass spectrometry",
    ),
    "CHMO:0001000": CvTerm(
        source="CHMO",
        accession="CHMO:0001000",
        name="chromatography",
    ),
    "MSIO:0000171": CvTerm(
        source="MSIO", accession="MSIO:0000171", name="mass spectrometry instrument"
    ),
}



from __future__ import annotations

import logging
from typing import Any

import jsonschema
import jsonschema.protocols

from mhd_model.model.v0_1.announcement.validation.base import ProfileValidator
from mhd_model.model.v0_1.announcement.validation.definitions import (
    CheckChildCvTermKeyValues,
    CheckCvTermKeyValue,
    CheckCvTermKeyValues,
)
from mhd_model.shared.exceptions import MhdValidationError
from mhd_model.shared.model import ProfileEnabledDataset
from mhd_model.shared.validation.definitions import (
    AccessibleCompactURI,
    AccessibleURI,
    AllowAnyCvTerm,
    AllowedChildrenCvTerms,
    AllowedCvList,
    AllowedCvTerms,
    ProfileValidationGroup,
)
from mhd_model.shared.validation.registry import (
    register_validator_class,
    unregister_validator_class,
)

logger = logging.getLogger(__name__)


class MhdAnnouncementFileValidator:
    validators = {
        "validation-group": ProfileValidationGroup,
        "allowed-cv-terms": AllowedCvTerms,
        "allowed-children-cv-terms": AllowedChildrenCvTerms,
        "allow-any-cv-term": AllowAnyCvTerm,
        "allowed-cv-list": AllowedCvList,
        "accessible-uri": AccessibleURI,
        "accessible-compact-uri": AccessibleCompactURI,
        "check-protocol-parameters": CheckChildCvTermKeyValues,
        "check-cv-term-key-values": CheckCvTermKeyValues,
        "check-cv-term-key-value": CheckCvTermKeyValue,
    }

    def register_validators(self):
        for k, v in self.validators.items():
            register_validator_class(k, v)

    def unregister_validators(self):
        for k, v in self.validators.items():
            unregister_validator_class(k)

    def validate(
        self, announcement_file_json: dict[str, Any]
    ) -> list[jsonschema.ValidationError]:
        self.register_validators()
        profile: ProfileEnabledDataset = ProfileEnabledDataset.model_validate(
            announcement_file_json
        )
        validator: jsonschema.protocols.Validator = ProfileValidator.new_instance(
            profile.schema_name, profile.profile_uri
        )
        if not validator:
            logger.error(
                "No validator found for schema %s with profile URI %s",
                profile.schema_name,
                profile.profile_uri,
            )
            raise MhdValidationError(
                f"No validator found for schema {profile.schema_name} with profile URI {profile.profile_uri}"
            )
        validations = validator.iter_errors(announcement_file_json)

        all_errors = [x for x in validations]

        def update_context(
            error: jsonschema.ValidationError, parent: jsonschema.ValidationError
        ):
            error.parent = parent
            if error.context:
                null_validators = []
                for item in error.context:
                    item.parent = error
                    update_context(item, error)
                item.context = [x for x in item.context if x not in null_validators]

        # profile_validator = ProfileValidator()

        for x in all_errors:
            # if x.validator in profile_validator.validators:
            update_context(x, x.parent)

        return all_errors



import warnings
from typing import Annotated, Any, Literal

from pydantic import AnyUrl, Field, model_validator

from mhd_model.model.v0_1.dataset.validation.profile.base import (
    EmbeddedRefValidation,
    RelationshipValidation,
)
from mhd_model.shared.model import CvTerm, MhdConfigModel
from mhd_model.shared.validation.definitions import (
    AllowAnyCvTerm,
    AllowedChildrenCvTerms,
    AllowedCvList,
    AllowedCvTerms,
)

warnings.filterwarnings("ignore", category=UserWarning)


class FilterCondition(MhdConfigModel):
    name: Annotated[str, Field()]
    relationship_name: Annotated[str, Field()]
    start_node_type: Annotated[None | str, Field()]
    expression: Annotated[None | str, Field()] = None
    expression_value: Annotated[None | str | CvTerm, Field()] = None


NodePropertyType = Literal["int", "str", "float", "CvTerm", "CvTermValue"]


class PropertyConstraint(MhdConfigModel):
    min_length: None | int = None
    max_length: None | int = None
    null_allowed: None | bool = None
    required: None | bool = None
    pattern: None | str = None
    allowed_types: None | NodePropertyType | list[NodePropertyType] = None

    def __str__(self) -> str:
        min = "Min Length: " + str(self.min_length) if self.min_length else ""
        max = "Max Length: " + str(self.max_length) if self.max_length else ""
        required = "Required" if self.required else ""
        return ", ".join(x for x in [min, max, required] if x)


class CvTermValidation(MhdConfigModel):
    min_count: Annotated[int, Field()] = 0
    node_type: Annotated[str, Field()]
    node_property_name: Annotated[None | str, Field()] = None
    validation: Annotated[
        AllowedCvTerms | AllowedChildrenCvTerms | AllowAnyCvTerm | AllowedCvList,
        Field(),
    ]
    condition: Annotated[None | list[FilterCondition], Field()] = None


class NodePropertyValidation(MhdConfigModel):
    node_type: Annotated[str, Field()]
    node_property_name: Annotated[None | str, Field()] = None
    contraints: Annotated[PropertyConstraint, Field()]


class NodeValidation(MhdConfigModel):
    node_type: Annotated[str, Field()]
    min: Annotated[int, Field()]
    max: Annotated[None | int, Field()] = None
    validations: Annotated[
        None | list[CvTermValidation | EmbeddedRefValidation | NodePropertyValidation],
        Field(),
    ] = None

    relationships: Annotated[None | list[RelationshipValidation], Field()] = None

    @model_validator(mode="wrap")
    @classmethod
    def validate_model(cls, v: Any, handler: callable) -> "NodeValidation":
        item: NodeValidation = handler(v)

        item.node_type = item.node_type.lower().replace(" ", "-")
        return item


class CvNodeValidation(NodeValidation):
    has_value: Annotated[bool, Field()] = False
    value_required: Annotated[bool, Field()] = False


class MhDatasetValidation(MhdConfigModel):
    mhd_nodes: Annotated[list[NodeValidation], Field()] = []
    cv_nodes: Annotated[list[CvNodeValidation], Field()] = []
    schema: Annotated[AnyUrl, Field()]



from pydantic import Field
from typing_extensions import Annotated

from mhd_model.model.v0_1.dataset.profiles.base.base import (
    BaseMhdRelationship,
    MhdObjectType,
)


class Relationship(BaseMhdRelationship):
    type_: Annotated[
        None | MhdObjectType,
        Field(
            frozen=True,
            alias="type",
            description="The type property identifies type of the MHD Relationship Object",
        ),
    ] = "relationship"



import datetime
import uuid
from typing import Any

from pydantic import AnyUrl, Field, field_validator, model_validator
from typing_extensions import Annotated

from mhd_model.shared.model import (
    CvTerm,
    CvTermValue,
    MhdConfigModel,
    QuantitativeValue,
)

NAMESPACE_VALUE = uuid.UUID("efb4f8e4-d08b-4979-916e-600c4985e7f2")
base_suffix = (
    r"[-a-zA-Z0-9]+--[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$"
)
OBJECT_UUID_PATTERN = rf"^mhd--{base_suffix}"
CV_TERM_UUID_PATTERN = rf"^cv--{base_suffix}"
CV_TERM_VALUE_UUID_PATTERN = rf"^cv-value--{base_suffix}"
RELATIONSHIP_UUID_PATTERN = rf"^rel--{base_suffix}"
OBJECT_TYPE_PATTERN = r"^[a-z0-9]+(?:-[a-z0-9]+)*$"


MhdObjectType = Annotated[str, Field(..., pattern=OBJECT_TYPE_PATTERN)]
setattr(MhdObjectType, "__name__", "MhdObjectType")

# MhdObjectType = Annotated[MhdObjectType, Field()]
MhdObjectId = Annotated[str, Field(pattern=OBJECT_UUID_PATTERN)]
setattr(MhdObjectId, "__name__", "MhdObjectId")
CvTermObjectId = Annotated[str, Field(pattern=CV_TERM_UUID_PATTERN)]
setattr(CvTermObjectId, "__name__", "CvTermObjectId")
CvTermValueObjectId = Annotated[str, Field(pattern=CV_TERM_VALUE_UUID_PATTERN)]
setattr(CvTermValueObjectId, "__name__", "CvTermValueObjectId")
MhdRelationshipObjectId = Annotated[str, Field(pattern=RELATIONSHIP_UUID_PATTERN)]
setattr(MhdRelationshipObjectId, "__name__", "MhdRelationshipObjectId")


class KeyValue(MhdConfigModel):
    key: None | MhdObjectId | CvTermObjectId | str | CvTerm = None
    value: (
        None
        | MhdObjectId
        | CvTermObjectId
        | CvTermValueObjectId
        | str
        | int
        | datetime.datetime
        | bool
        | float
        | CvTerm
        | CvTermValue
        | QuantitativeValue
    ) = None


# class CvTermKeyValue(MhdConfigModel):
#     key: CvTerm
#     value: str | datetime.datetime | bool | CvTerm | CvTermValue | QuantitativeValue


class IdentifiableMhdModel(MhdConfigModel):
    id_: Annotated[
        None
        | MhdObjectId
        | CvTermObjectId
        | CvTermValueObjectId
        | MhdRelationshipObjectId,
        Field(
            alias="id",
            description="Unique identifier of graph node",
        ),
    ] = None
    type_: Annotated[
        None | MhdObjectType,
        Field(
            alias="type",
            description="The type property identifies the type of MHD Object. "
            "Its value MUST be the name of one of the types of MHD Objects",
        ),
    ]


class BaseMhdModel(IdentifiableMhdModel):
    id_: Annotated[
        None | MhdObjectId,
        Field(
            alias="id", description="The id property uniquely identifies the object."
        ),
    ] = None
    type_: Annotated[None | MhdObjectType, Field(frozen=False, alias="type")] = (
        "base-mhd-model"
    )
    created_by_ref: Annotated[
        None | CvTermValueObjectId,
        Field(
            description="The id property of the data-provider who created the object.",
        ),
    ] = None
    tag_list: Annotated[
        None | list[KeyValue],
        Field(description="Key-value tags related to the object."),
    ] = None
    external_reference_list: Annotated[
        None | list[KeyValue],
        Field(description="External references related to the object."),
    ] = None
    url_list: Annotated[
        None | list[AnyUrl],
        Field(description="URL list related to the object."),
    ] = None

    @field_validator("id_", mode="before")
    @classmethod
    def id_validator(cls, v) -> str:
        if isinstance(v, str):
            try:
                uuid.UUID(v.split("--")[2])
                return v
            except Exception:
                raise ValueError(f"invalid string structure {v}")
        raise ValueError("invalid type")

    @model_validator(mode="wrap")
    @classmethod
    def validate_model(cls, v: Any, handler) -> "BaseMhdModel":
        item: BaseMhdModel = handler(v)

        if item.type_ and not item.id_:
            item.id_ = f"mhd--{item.type_}--{uuid.uuid4()}"
        return item

    def get_unique_id(self):
        return self.id_

    def __hash__(self) -> int:
        return hash(self.get_unique_id())


class BaseLabeledMhdModel(BaseMhdModel):
    label: Annotated[None | str, Field(exclude=True)] = None

    @model_validator(mode="wrap")
    @classmethod
    def validate_model(cls, v: Any, handler) -> "BaseLabeledMhdModel":
        item: BaseLabeledMhdModel = handler(v)

        if item.type_ and not item.id_:
            item.id_ = f"mhd--{item.type_}--{uuid.uuid4()}"
        if not item.label:
            item.label = item.get_label()
        return item

    def get_label(self) -> str:
        return self.id_ or ""


class Definition(BaseLabeledMhdModel):
    name: Annotated[None | str, Field()] = None
    definition_type: Annotated[None | CvTerm, Field()] = None

    @model_validator(mode="wrap")
    @classmethod
    def validate_model(cls, v: Any, handler) -> "Definition":
        item: Definition = handler(v)
        if item.type_ and not item.id_:
            str_repr = item.get_unique_id()
            identifier_name = f"{item.type_}--{str_repr}"
            identifier = str(uuid.uuid5(NAMESPACE_VALUE, name=identifier_name))
            item.id_ = f"{item.type_}--{identifier}"
        if not item.label:
            item.label = item.get_label()
        return item

    def get_unique_id(self):
        return f"{self.name or ''},{self.definition_type.get_unique_id() if self.definition_type else ''}"

    def __hash__(self) -> int:
        return hash(self.get_unique_id())

    def get_label(self):
        return self.name or ""


class DefinitionValue(KeyValue): ...


class BasicCvTermModel(CvTerm, IdentifiableMhdModel):
    id_: Annotated[
        None | CvTermObjectId,
        Field(
            ...,
            alias="id",
            description="The id property uniquely identifies the object.",
        ),
    ] = None
    label: Annotated[None | str, Field(exclude=True)] = None
    type_: Annotated[None | MhdObjectType, Field(..., alias="type")]

    @field_validator("id_", mode="before")
    @classmethod
    def id_validator(cls, v) -> str:
        if isinstance(v, str):
            try:
                uuid.UUID(v.split("--")[2])
                return v
            except Exception:
                raise ValueError(f"invalid string structure {v}")
        raise ValueError("invalid type")

    @model_validator(mode="wrap")
    @classmethod
    def validate_model(cls, v: Any, handler) -> "BasicCvTermModel":
        item: BasicCvTermModel = handler(v)
        if item.type_ and not item.id_:
            str_repr = item.get_unique_id()
            identifier_name = f"{item.type_}--{str_repr}"
            identifier = str(uuid.uuid5(NAMESPACE_VALUE, name=identifier_name))
            item.id_ = f"cv--{item.type_}--{identifier}"
        if not item.label:
            item.label = item.get_label()
        return item

    def get_label(self):
        return self.name or self.id_ or ""

    def __hash__(self) -> int:
        return hash(self.get_unique_id())


class BasicCvTermValueModel(CvTermValue, IdentifiableMhdModel):
    id_: Annotated[
        None | CvTermValueObjectId,
        Field(
            ...,
            alias="id",
            description="The id property uniquely identifies the object.",
        ),
    ] = None
    label: Annotated[None | str, Field(exclude=True)] = None
    type_: Annotated[MhdObjectType, Field(frozen=False, alias="type")] = "cv-term-value"

    @model_validator(mode="wrap")
    @classmethod
    def validate_model(cls, v: Any, handler) -> "BasicCvTermValueModel":
        item: BasicCvTermValueModel = handler(v)
        if item.type_ and not item.id_:
            str_repr = item.get_unique_id()
            identifier_name = f"{item.type_}--{str_repr}"
            identifier = str(uuid.uuid5(NAMESPACE_VALUE, name=identifier_name))
            item.id_ = f"cv-value--{item.type_}--{identifier}"
        if not item.label:
            item.label = item.get_label()
        return item

    def get_label(self):
        return self.value or self.name or self.id_ or ""

    def __hash__(self) -> int:
        return hash(self.get_unique_id())


class BaseMhdRelationship(BaseMhdModel):
    id_: Annotated[None | MhdRelationshipObjectId, Field(alias="id")] = None
    source_ref: MhdObjectId | CvTermObjectId | CvTermValueObjectId
    relationship_name: str
    target_ref: MhdObjectId | CvTermObjectId | CvTermValueObjectId
    source_role: None | str = None
    target_role: None | str = None

    @model_validator(mode="wrap")
    @classmethod
    def validate_model(cls, v: Any, handler) -> "BaseMhdRelationship":
        item: BaseMhdRelationship = handler(v)
        if item.type_ and not item.id_:
            str_repr = item.get_unique_id()
            identifier_name = f"{item.type_}--{str_repr}"
            identifier = str(uuid.uuid5(NAMESPACE_VALUE, name=identifier_name))
            item.id_ = f"rel--{item.type_}--{identifier}"

        return item

    def get_unique_id(self):
        return f"{self.source_ref or ''},{self.relationship_name or ''},{self.target_ref or ''}"

    def get_label(self):
        return self.relationship_name



import click

from mhd_model.convertors.sdrf.mhd2sdrf import create_sdrf_files


@click.command(name="sdrf", no_args_is_help=True)
@click.option(
    "--output-dir",
    default="outputs",
    help="Output directory for SDRF files",
)
@click.option(
    "--output-filename",
    default=None,
    help="SDRF filename (e.g., MHD000001_assay1.sdrf.tsv)",
)
@click.option(
    "--assay-name",
    default=None,
    help="Name of assay. If it is not defined, SDRF files will be created for all assays.",
)
@click.argument("mhd_study_id")
@click.argument("mhd_model_file_path")
def create_sdrf_file_task(
    mhd_study_id: str,
    mhd_model_file_path: str,
    assay_name: None | str,
    output_dir: None | str,
    output_filename: None | str,
):
    """Create SDRF file from MHD data model file.

    Args:

    mhd_study_id (str): MHD study identifier

    mhd_model_file_path (str): MHD data model path

    assay_name (str): Name of assay.
    If it is not defined, SDRF files will be created for all assays

    output_dir (str): Output directory of SDRF files.

    output_filename (str): Name of SDRF file. It is applicable if
    """
    try:
        sdrf_files = create_sdrf_files(
            mhd_model_file_path,
            output_dir,
            assay_name=assay_name,
            sdrf_output_filename=output_filename,
        )
        if sdrf_files:
            click.echo(f"{mhd_study_id}: {len(sdrf_files)} SDRF files created.")
        else:
            click.echo(f"There is no SDRF file for {mhd_study_id}.")
            exit(1)
    except Exception as ex:
        click.echo(f"{mhd_study_id} SDRF file creation failed. {str(ex)}")
        exit(1)
